MODULE traadv_eiv_tam
#if defined key_tam
   !!======================================================================
   !!                    ***  MODULE  traadv_eiv_tam  ***
   !! Ocean active tracers:  advection trend - eddy induced velocity
   !!                        Tangent and Adjoint Module
   !!======================================================================
   !! History of the direct module:
   !!         9.0  !  05-11  (G. Madec)  Original code, from traldf and zdf _iso
   !! History of the T&A module:
   !!         9.0  !  08-11  (A. Vidard) skeleton
   !!----------------------------------------------------------------------
# if defined key_traldf_eiv   ||   defined key_esopa
   !!----------------------------------------------------------------------
   !!   'key_traldf_eiv'                  rotation of the lateral mixing tensor
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!   tra_ldf_iso : update the tracer trend with the horizontal component
   !!                 of iso neutral laplacian operator or horizontal 
   !!                 laplacian operator in s-coordinate
   !!----------------------------------------------------------------------
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpi, jpj, jpk 
   USE in_out_manager, ONLY: & ! I/O manager 
      & lwp, numout              

   IMPLICIT NONE
   PRIVATE

   PUBLIC   tra_adv_eiv_tan    ! routine called by traadv_tam.F90
   PUBLIC   tra_adv_eiv_adj    ! routine called by traadv_tam.F90
!!   PUBLIC   tra_adv_eiv_adj_tst! routine called by traadv_tam.F90

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "ldftra_substitute.h90"
#  include "ldfeiv_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !!  OPA 9.0 , LOCEAN-IPSL (2006) 
   !! $Id: traadv_eiv.F90 1152 2008-06-26 14:11:13Z rblod $ 
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE tra_adv_eiv_tan( kt, puntl, pvntl, pwntl )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_adv_eiv_tan  ***
      !! 
      !! ** Purpose of the direct routine:
      !!      Compute the before horizontal tracer (t & s) diffusive 
      !!      trend and add it to the general trend of tracer equation.
      !!
      !! ** Method of the direct routine:
      !!      The eddy induced advection is computed from the slope
      !!      of iso-neutral surfaces computed in routine ldf_slp as follows:
      !!         zu_eiv =  1/(e2u e3u)   dk[ aeiu e2u mi(wslpi) ]
      !!         zv_eiv =  1/(e1v e3v)   dk[ aeiv e1v mj(wslpj)
      !!         zw_eiv = -1/(e1t e2t) { di[ aeiu e2u mi(wslpi) ]
      !!                               + dj[ aeiv e1v mj(wslpj) ] }
      !!      add the eiv component to the model velocity:
      !!         p.n = p.n + z._eiv
      !!
      !! ** Action  : - add to p.n the eiv component
      !!----------------------------------------------------------------------
      INTEGER , INTENT(in   )                         ::   kt    ! ocean time-step index
      REAL(wp), INTENT(inout), DIMENSION(jpi,jpj,jpk) ::   puntl   ! in : 3 ocean velocity components 
      REAL(wp), INTENT(inout), DIMENSION(jpi,jpj,jpk) ::   pvntl   ! out: 3 ocean velocity components
      REAL(wp), INTENT(inout), DIMENSION(jpi,jpj,jpk) ::   pwntl   !      increased by the eiv
      !!
!!     IF ( ln_key_diaeiv) THEN
!!         IF (lwp) WRITE(numout,*) 'key_diaeiv not available yet in NEMOTAM'
!!         CALL abort
!!      ENDIF

   END SUBROUTINE tra_adv_eiv_tan
   SUBROUTINE tra_adv_eiv_adj( kt, punad, pvnad, pwnad )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_adv_eiv_adj  ***
      !! 
      !! ** Purpose of the direct routine:
      !!      Compute the before horizontal tracer (t & s) diffusive 
      !!      trend and add it to the general trend of tracer equation.
      !!
      !! ** Method of the direct routine:
      !!      The eddy induced advection is computed from the slope
      !!      of iso-neutral surfaces computed in routine ldf_slp as follows:
      !!         zu_eiv =  1/(e2u e3u)   dk[ aeiu e2u mi(wslpi) ]
      !!         zv_eiv =  1/(e1v e3v)   dk[ aeiv e1v mj(wslpj)
      !!         zw_eiv = -1/(e1t e2t) { di[ aeiu e2u mi(wslpi) ]
      !!                               + dj[ aeiv e1v mj(wslpj) ] }
      !!      add the eiv component to the model velocity:
      !!         p.n = p.n + z._eiv
      !!
      !! ** Action  : - add to p.n the eiv component
      !!----------------------------------------------------------------------
      INTEGER , INTENT(in   )                         ::   kt    ! ocean time-step index
      REAL(wp), INTENT(inout), DIMENSION(jpi,jpj,jpk) ::   punad   ! in : 3 ocean velocity components 
      REAL(wp), INTENT(inout), DIMENSION(jpi,jpj,jpk) ::   pvnad   ! out: 3 ocean velocity components
      REAL(wp), INTENT(inout), DIMENSION(jpi,jpj,jpk) ::   pwnad   !      increased by the eiv
      !!

   END SUBROUTINE tra_adv_eiv_adj

#else
   !!----------------------------------------------------------------------
   !!   Dummy module :             No rotation of the lateral mixing tensor
   !!----------------------------------------------------------------------
CONTAINS
   SUBROUTINE tra_adv_eiv_tan( kt, puntl, pvntl, pwntl ) ! Empty routine
      REAL, DIMENSION(:,:,:) ::   puntl, pvntl, pwntl
      WRITE(*,*) 'tra_adv_eiv_tam: You should not have seen this print! error?', kt, puntl(1,1,1), pvntl(1,1,1), pwntl(1,1,1)
   END SUBROUTINE tra_adv_eiv_tan
   SUBROUTINE tra_adv_eiv_adj( kt, punad, pvnad, pwnad ) ! Empty routine
      REAL, DIMENSION(:,:,:) ::   punad, pvnad, pwnad
      WRITE(*,*) 'tra_adv_eiv_tam: You should not have seen this print! error?', kt, punad(1,1,1), pvnad(1,1,1), pwnad(1,1,1)
   END SUBROUTINE tra_adv_eiv_adj
# endif
#endif
   !!==============================================================================
END MODULE traadv_eiv_tam
