MODULE trazdf_tam
#ifdef key_tam
   !!==============================================================================
   !!                 ***  MODULE  trazdf_zdf  ***
   !! Ocean active tracers:  vertical component of the tracer mixing trend
   !!                 Tangent and Adjoint Module
   !!==============================================================================
   !! History of the direct module:  
   !!            9.0  !  05-11  (G. Madec)  Original code
   !! History of the TAM module:
   !!            9.0  !  08-06  (A. Vidard) Skeleton
   !!            9.0  !  09-01  (A. Vidard) TAM of the 05-11 version
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   tra_zdf_tan  : Update the tracer trend with the vertical diffusion (tan)
   !!   tra_zdf_adj  : Update the tracer trend with the vertical diffusion (adj)
   !!       zdf_ctl  : ???
   !!----------------------------------------------------------------------
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpk,                 &
      & lk_esopa
   USE oce_tam       , ONLY: & ! ocean dynamics and active tracers
      & ta_tl,               &
      & sa_tl,               &
      & ta_ad,               &
      & sa_ad
   USE dom_oce       , ONLY: & ! ocean space and time domain
      & tmask,               &
      & lk_vvl,              &
      & neuler,              &
      & rdttra,              &
      & ln_sco
   USE ldftra_oce    , ONLY: &
      & ln_traldf_iso,       &
      & ln_traldf_hor
   USE zdf_oce      , ONLY : &
      & ln_zdfexp
   USE trazdf_exp_tam, ONLY: & ! vertical diffusion: explicit (tra_zdf_exp     routine)
      & tra_zdf_exp_tan,     &
      & tra_zdf_exp_adj,     &
      & tra_zdf_exp_adj_tst
   USE trazdf_imp_tam, ONLY: & ! vertical diffusion: implicit (tra_zdf_imp     routine)
      & tra_zdf_imp_tan,     &
      & tra_zdf_imp_adj,     &
#if defined key_tst_tlm
      & tra_zdf_imp_tlm_tst, &
#endif
      & tra_zdf_imp_adj_tst
   USE in_out_manager, ONLY: & ! I/O manager 
      & lwp,                 &
      & numout,              & 
      & nitend,              & 
      & nit000,              &
      & ctl_stop
   USE prtctl        , ONLY: &
      & prt_ctl

   IMPLICIT NONE
   PRIVATE

   PUBLIC  &
      & tra_zdf_tan, &
      & tra_zdf_adj         ! routines called by step_tam.F90
   PUBLIC  tra_zdf_adj_tst  ! routine called by tst.F90
#if defined key_tst_tlm
   PUBLIC  tra_zdf_tlm_tst     ! routine called by tst.F90 
#endif
   INTEGER ::   nzdf = 0               ! type vertical diffusion algorithm used
      !                                ! defined from ln_zdf...  namlist logicals)

   REAL(wp), DIMENSION(jpk) ::   r2dt  ! vertical profile time-step, = 2 rdttra
      !                                ! except at nit000 (=rdttra) if neuler=0

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "zdfddm_substitute.h90"
#  include "vectopt_loop_substitute.h90"

CONTAINS
   
   SUBROUTINE tra_zdf_tan( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_zdf_tan  ***
      !!
      !! ** Purpose of the direct routine:   
      !!            compute the vertical ocean tracer physics.
      !!---------------------------------------------------------------------
      INTEGER, INTENT( in ) ::   kt      ! ocean time-step index

      !!---------------------------------------------------------------------

      IF( kt == nit000 )   CALL zdf_ctl_tam          ! initialisation & control of options

      !                                          ! set time step
      IF( neuler == 0 .AND. kt == nit000 ) THEN     ! at nit000
         r2dt(:) =  rdttra(:)                          ! = rdtra (restarting with Euler time stepping)
      ELSEIF( kt <= nit000 + 1) THEN                ! at nit000 or nit000+1
         r2dt(:) = 2. * rdttra(:)                      ! = 2 rdttra (leapfrog)
      ENDIF


      IF( lk_vvl ) THEN !CALL dom_vvl_ssh( kt )      ! compute ssha field
         IF (lwp) WRITE(numout,*) 'key_vvl not available in tangent yet'
         CALL abort
      END IF
         
      SELECT CASE ( nzdf )                       ! compute lateral mixing trend and add it to the general trend
      CASE ( -1 )                                       ! esopa: test all possibility with control print
         CALL tra_zdf_exp_tan    ( kt, r2dt )
         CALL prt_ctl( tab3d_1=ta_tl, clinfo1=' zdf0 - Ta_tl: ', mask1=tmask,               &
            &          tab3d_2=sa_tl, clinfo2=       ' Sa_tl: ', mask2=tmask, clinfo3='tra' )
         CALL tra_zdf_imp_tan    ( kt, r2dt )
         CALL prt_ctl( tab3d_1=ta_tl, clinfo1=' zdf1 - Ta_tl: ', mask1=tmask,               &
            &          tab3d_2=sa_tl, clinfo2=       ' Sa_tl: ', mask2=tmask, clinfo3='tra' )

      CASE ( 0 )                                       ! explicit scheme
         CALL tra_zdf_exp_tan    ( kt, r2dt )

      CASE ( 1 )                                       ! implicit scheme 
         CALL tra_zdf_imp_tan    ( kt, r2dt )

      END SELECT

   END SUBROUTINE tra_zdf_tan
   SUBROUTINE tra_zdf_adj( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_zdf_adj  ***
      !!
      !! ** Purpose of the direct routine:   
      !!            compute the vertical ocean tracer physics.
      !!---------------------------------------------------------------------
      INTEGER, INTENT( in ) ::   kt      ! ocean time-step index

      !!---------------------------------------------------------------------

      IF( kt == nitend )   CALL zdf_ctl_tam          ! initialisation & control of options
      !                                          ! set time step
      IF( neuler == 0 .AND. kt == nit000 ) THEN     ! at nit000
         r2dt(:) =  rdttra(:)                          ! = rdtra (restarting with Euler time stepping)
      ELSEIF( kt <= nit000 + 1) THEN                ! at nit000 or nit000+1
         r2dt(:) = 2. * rdttra(:)                      ! = 2 rdttra (leapfrog)
      ELSEIF( kt == nitend) THEN
         r2dt(:) = 2. * rdttra(:)
      ENDIF
         
      SELECT CASE ( nzdf )                       ! compute lateral mixing trend and add it to the general trend
      CASE ( -1 )                                       ! esopa: test all possibility with control print
         CALL tra_zdf_exp_adj    ( kt, r2dt )
         CALL prt_ctl( tab3d_1=ta_ad, clinfo1=' zdf0 - Ta_ad: ', mask1=tmask,               &
            &          tab3d_2=sa_ad, clinfo2=       ' Sa_ad: ', mask2=tmask, clinfo3='tra' )
         CALL tra_zdf_imp_adj    ( kt, r2dt )
         CALL prt_ctl( tab3d_1=ta_ad, clinfo1=' zdf1 - Ta_ad: ', mask1=tmask,               &
            &          tab3d_2=sa_ad, clinfo2=       ' Sa_ad: ', mask2=tmask, clinfo3='tra' )

      CASE ( 0 )                                       ! explicit scheme
         CALL tra_zdf_exp_adj    ( kt, r2dt )

      CASE ( 1 )                                       ! implicit scheme 
         CALL tra_zdf_imp_adj    ( kt, r2dt )

      END SELECT


      IF( lk_vvl ) THEN !CALL dom_vvl_ssh( kt )      ! compute ssha field
         IF (lwp) WRITE(numout,*) 'key_vvl not available in tangent yet'
         CALL abort
      END IF

   END SUBROUTINE tra_zdf_adj
   SUBROUTINE tra_zdf_adj_tst( kumadt )
      !! ** Purpose : Test the adjoint routines.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !!       
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      !! * Local declarations
      ! init
      CALL zdf_ctl_tam 
      ! Test the explicit formulation            
      CALL tra_zdf_exp_adj_tst    ( kumadt )
      ! Test the implicit formulation            
      CALL tra_zdf_imp_adj_tst    ( kumadt )
   END SUBROUTINE tra_zdf_adj_tst
   !!==============================================================================
   SUBROUTINE zdf_ctl_tam
      !!----------------------------------------------------------------------
      !!                 ***  ROUTINE zdf_ctl_tam  ***
      !!
      !! ** Purpose :   Choose the vertical mixing scheme
      !!
      !! ** Method  :   Set nzdf from ln_zdfexp
      !!      nzdf = 0   explicit (time-splitting) scheme (ln_zdfexp=T)
      !!           = 1   implicit (euler backward) scheme (ln_zdfexp=F)
      !!      NB: rotation of lateral mixing operator or TKE or KPP scheme,
      !!      the implicit scheme is required.
      !!----------------------------------------------------------------------
      USE zdftke
      USE zdfkpp
      !!----------------------------------------------------------------------

      !  Define the vertical tracer physics scheme
      ! ==========================================

      ! Choice from ln_zdfexp already read in namelist in zdfini module
      IF( ln_zdfexp ) THEN               ! use explicit scheme
         nzdf = 0
      ELSE                               ! use implicit scheme
         nzdf = 1
      ENDIF

      ! Force implicit schemes
      IF( lk_zdftke .OR. lk_zdfkpp   )   nzdf = 1      ! TKE or KPP physics
      IF( ln_traldf_iso              )   nzdf = 1      ! iso-neutral lateral physics
      IF( ln_traldf_hor .AND. ln_sco )   nzdf = 1      ! horizontal lateral physics in s-coordinate

      IF( ln_zdfexp .AND. nzdf == 1 )   THEN
         CALL ctl_stop( 'tra_zdf_tam : If using the rotation of lateral mixing operator or TKE ', &
            &           '            or KPP scheme, the implicit scheme is required, set ln_zdfexp = .false.' )
      ENDIF

      ! Test: esopa
      IF( lk_esopa )    nzdf = -1                      ! All schemes used

      IF(lwp) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'tra:zdf_ctl_tam : vertical tracer physics scheme'
         WRITE(numout,*) '~~~~~~~~~~~~~~~'
         IF( nzdf == -1 )   WRITE(numout,*) '              ESOPA test All scheme used'
         IF( nzdf ==  0 )   WRITE(numout,*) '              Explicit time-splitting scheme'
         IF( nzdf ==  1 )   WRITE(numout,*) '              Implicit (euler backward) scheme'
      ENDIF

   END SUBROUTINE zdf_ctl_tam
#if defined key_tst_tlm
   SUBROUTINE tra_zdf_tlm_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE tra_zdf_tlm_tst ***
      !!
      !! ** Purpose : Test the tangent routine.
      !!
      !! ** Method  : Verify the tangent with Taylor expansion 
      !!           
      !!                 M(x+hdx) = M(x) + L(hdx) + O(h^2)
      !!
      !!              where  L   = tangent routine
      !!                     M   = direct routine
      !!                     dx  = input perturbation (random field)
      !!                     h   = ration on perturbation 
      !!                   
      !! History :
      !!        ! 09-08 (A. Vigilant)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      !! * Local declarations
      ! init
      CALL zdf_ctl_tam 
      ! Test the explicit formulation            
!      CALL tra_zdf_exp_adj_tst    ( kumadt )
      ! Test the implicit formulation            
      CALL tra_zdf_imp_tlm_tst    ( kumadt )
   END SUBROUTINE tra_zdf_tlm_tst
#endif
#endif
END MODULE trazdf_tam
