MODULE istate_tam
#ifdef key_tam
   !!======================================================================
   !!                     ***  MODULE  istate_tam  ***
   !! Ocean state   :  initial state setting
   !!                  Tangent and Adjoint Module
   !!=====================================================================
   !! History of the direct module:
   !!             4.0  !  89-12  (P. Andrich)  Original code
   !!             5.0  !  91-11  (G. Madec)  rewritting
   !!             6.0  !  96-01  (G. Madec)  terrain following coordinates
   !!             8.0  !  01-09  (M. Levy, M. Ben Jelloul)  istate_eel
   !!             8.0  !  01-09  (M. Levy, M. Ben Jelloul)  istate_uvg
   !!             9.0  !  03-08  (G. Madec)  F90: Free form, modules
   !!             9.0  !  03-09  (G. Madec, C. Talandier)  add EEL R5
   !!             9.0  !  04-05  (A. Koch-Larrouy)  istate_gyre 
   !!             9.0  !  06-07  (S. Masson)  distributed restart using iom
   !! History of the T&A module:
   !!             9.0  !  09-04  (F. Vigilant) TAM of the 06-07 version
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   istate_init_tan : initial state setting for the tangent model
   !!----------------------------------------------------------------------
   USE par_kind      , ONLY: & ! Precision variables
      & wp
  USE par_oce        , ONLY: & ! Ocean space and time domain variables
      & jpi, jpj, jpk, jpiglo
   USE oce_tam ! ocean dynamics and active tracers 
   USE oce           , ONLY: & ! ocean dynamics and active tracers 
      & tb, sb
   USE dom_oce       , ONLY: & ! ocean space and time domain 
      & neuler, ln_zps, mig, &
      & mjg , nldi, nldj,    &
      & nlei, nlej, lk_vvl,  &
      & e1t, e2t, e1u, e2u,  &
      & e1v, e2v,            &
# if defined key_vvl
      & e3t_1,               &
# else
#  if defined key_zco
      & e3t_0, e3w_0,        &
#  else
      & e3u, e3v, e3t, e3w,  &
#  endif
# endif
      & tmask, umask, vmask, &
      & n_cla
   USE daymod        , ONLY: &
      & day_init
   USE c1d           , ONLY: &
      & lk_c1d
   USE restart       , ONLY: & ! ocean restart
      & numror             
   USE in_out_manager, ONLY: & ! I/O manager
      & numout, lwp, nit000
   USE wzvmod_tam    , ONLY: & ! verctical velocity               (wzv     routine)
      & wzv_tan, wzv_adj
   USE zpshde_tam    , ONLY: & ! partial step: hor. derivative (zps_hde routine)
      & zps_hde_tan,         &
      & zps_hde_adj
   USE eosbn2_tam    , ONLY: &
      & eos_tan, eos_adj
   USE divcur_tam    , ONLY: &
      & div_cur_tan, div_cur_adj
   USE cla_div_tam    , ONLY: &
      & div_cla_tan, div_cla_adj
   USE tstool_tam    , ONLY: &
      & prntst_adj,          &
      & stds  , stdt, stdr,  &
      & stdssh, stdu, stdv
   USE gridrandom    , ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld,     ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE paresp        , ONLY: & ! Normalized energy weights            
      & wesp_ssh

   IMPLICIT NONE
   PRIVATE

   PUBLIC   istate_init_tan       ! routine called by step.F90
   PUBLIC   istate_init_adj       ! routine called by step.F90
   PUBLIC   istate_init_adj_tst   ! routine called by tst.F90
    
   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !!   OPA 9.0 , LOCEAN-IPSL (2006) 
   !! $Id: istate.F90 1200 2008-09-24 13:05:20Z rblod $
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE istate_init_tan
      !!----------------------------------------------------------------------
      !!                   ***  ROUTINE istate_init Tangent ***
      !! 
      !! ** Purpose :   Initialization of the dynamics and tracer fields.
      !!----------------------------------------------------------------------
      USE eosbn2_tam          ! eq. of state, Brunt Vaisala frequency (eos     routine)

      IF(lwp) WRITE(numout,*)
      IF(lwp) WRITE(numout,*) 'istate_ini_tan : Initialization of the dynamics and tracers'
      IF(lwp) WRITE(numout,*) '~~~~~~~~~~'

      rhd_tl  (:,:,:) = 0.e0
      rhop_tl (:,:,:) = 0.e0
      rn2_tl  (:,:,:) = 0.e0 

      !                                    ! Start from rest
      !                                    ! ---------------
      numror = 0                              ! define numror = 0 -> no restart file to read
      neuler = 0                              ! Set time-step indicator at nit000 (euler forward)
      CALL day_init                           ! model calendar (using both namelist and restart infos)
      !                                       ! Initialization of ocean to zero

      !     before fields       !       now fields          
      ub_tl   (:,:,:) = un_tl   (:,:,:)    
      vb_tl   (:,:,:) = vn_tl   (:,:,:)    
      tb_tl   (:,:,:) = tn_tl   (:,:,:)
      sb_tl   (:,:,:) = sn_tl   (:,:,:)
      sshb_tl (  :,:) = sshn_tl (  :,:)
      !
      rotb_tl (:,:,:) = rotn_tl (:,:,:)   ! Update before fields
      hdivb_tl(:,:,:) = hdivn_tl(:,:,:)  


      CALL eos_tan( tb, sb, tb_tl, sb_tl, rhd_tl, rhop_tl )        ! before potential and in situ densities
         
      IF( ln_zps .AND. .NOT. lk_c1d )   &
            &             CALL zps_hde_tan( nit000, tb, sb, tb_tl, sb_tl, rhd_tl,  &  ! Partial steps: before Horizontal DErivative
            &                                  gtu_tl, gsu_tl, gru_tl,             &  ! of t, s, rd at the bottom ocean level
            &                                  gtv_tl, gsv_tl, grv_tl )
         



      !                                       ! Vertical velocity
      !                                       ! -----------------

      IF( .NOT. lk_vvl )    CALL wzv_tan( nit000 )                         ! from horizontal divergence
      !
   END SUBROUTINE istate_init_tan

   SUBROUTINE istate_init_adj
      !!----------------------------------------------------------------------
      !!                   ***  ROUTINE istate_init  Adjoint Module ***
      !! 
      !! ** Purpose :   Initialization of the dynamics and tracer fields.
      !!----------------------------------------------------------------------
      !! * Modules used
      USE eosbn2_tam          ! eq. of state, Brunt Vaisala frequency (eos     routine)

      !! * Local declarations
      INTEGER ::             &
         & ji,               &  ! dummy loop indices
         & jj,               &        
         & jk

      IF(lwp) WRITE(numout,*)
      IF(lwp) WRITE(numout,*) 'istate_ini_adj : Initialization of the dynamics and tracers'
      IF(lwp) WRITE(numout,*) '~~~~~~~~~~'


      !                                       ! Vertical velocity
      !                                       ! -----------------

      IF( .NOT. lk_vvl )    CALL wzv_adj( nit000 )                         ! from horizontal divergence

      IF( ln_zps .AND. .NOT. lk_c1d )   &
            &             CALL zps_hde_adj( nit000, tb, sb, tb_ad, sb_ad, rhd_ad,  &  ! Partial steps: before Horizontal DErivative
            &                                  gtu_ad, gsu_ad, gru_ad,             &  ! of t, s, rd at the bottom ocean level
            &                                  gtv_ad, gsv_ad, grv_ad )
         

      CALL eos_adj( tb, sb, tb_ad, sb_ad, rhd_ad, rhop_ad )        ! before potential and in situ densities

      !     before fields       !       now fields          
      DO jk = 1, jpk
         DO jj = 1, jpj
            DO ji = 1, jpi         
               un_ad   (ji, jj, jk) = un_ad(ji, jj, jk) + ub_ad (ji, jj, jk)
               ub_ad   (ji, jj, jk) = 0.0_wp
            END DO
         END DO   
      END DO
      DO jk = 1, jpk
         DO jj = 1, jpj
            DO ji = 1, jpi         
               vn_ad   (ji, jj, jk) = vn_ad(ji, jj, jk) + vb_ad (ji, jj, jk)
               vb_ad   (ji, jj, jk) = 0.0_wp
            END DO
         END DO   
      END DO
      DO jk = 1, jpk
         DO jj = 1, jpj
            DO ji = 1, jpi         
               tn_ad   (ji, jj, jk) = tn_ad(ji, jj, jk) + tb_ad (ji, jj, jk)
               tb_ad   (ji, jj, jk) = 0.0_wp
            END DO
         END DO   
      END DO
      DO jk = 1, jpk
         DO jj = 1, jpj
            DO ji = 1, jpi         
               sn_ad   (ji, jj, jk) = sn_ad(ji, jj, jk) + sb_ad (ji, jj, jk)
               sb_ad   (ji, jj, jk) = 0.0_wp
            END DO
         END DO   
      END DO
      DO jj = 1, jpj
         DO ji = 1, jpi         
             sshn_ad   (ji, jj    ) = sshn_ad(ji, jj  ) + sshb_ad (ji, jj  )
             sshb_ad   (ji, jj    ) = 0.0_wp
         END DO   
      END DO
      DO jk = 1, jpk
         DO jj = 1, jpj
            DO ji = 1, jpi         
               rotn_ad (ji, jj, jk) = rotn_ad(ji, jj, jk) + rotb_ad (ji, jj, jk)
               rotb_ad (ji, jj, jk) = 0.0_wp
            END DO
         END DO   
      END DO
      DO jk = 1, jpk
         DO jj = 1, jpj
            DO ji = 1, jpi         
               hdivn_ad(ji, jj, jk) = hdivn_ad(ji, jj, jk) + hdivb_ad (ji, jj, jk)
               hdivb_ad(ji, jj, jk) = 0.0_wp
            END DO
         END DO   
      END DO
      !


      rhd_ad  (:,:,:) = 0.0_wp
      rhop_ad (:,:,:) = 0.0_wp
      rn2_ad  (:,:,:) = 0.0_wp 

      !
   END SUBROUTINE istate_init_adj

   SUBROUTINE istate_init_adj_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE istate_init_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !!                   
      !! History :
      !!        ! 09-05 (F. Vigilant)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit

      INTEGER ::         &
         & ji,           &        ! dummy loop indices
         & jj,           &        
         & jk,           & 
         & kt
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator

      !! * Local declarations
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & ztn_tlin,     & ! Tangent input: temperature
         & zsn_tlin,     & ! Tangent input: salinity
         & zun_tlin,     & ! Tangent input: velocity
         & zvn_tlin,     & ! Tangent input: velocity
         & zrd_tlin,     & ! Tangent input:
         & ztn_adout,    & ! Adjoint output: temperature
         & zsn_adout,    & ! Adjoint output: salinity
         & zun_adout,    & ! Adjoint input: velocity
         & zvn_adout,    & ! Adjoint input: velocity
         & ztb_tlout,    & ! Tangent output: temperature
         & zsb_tlout,    & ! Tangent output: salinity
         & zub_tlout,    & ! Tangent output: velocity
         & zvb_tlout,    & ! Tangent output: velocity 
         & zrd_tlout,    & ! Tangent output: 
         & zrhop_tlout,  & ! Tangent output:
         & zrd_adin,     & ! Adjoint input:
         & zrhop_adin,   & ! Adjoint input:
         & ztb_adin,     & ! Adjoint input: temperature
         & zsb_adin,     & ! Adjoint input: salinity
         & zub_adin,     & ! Adjoint input: velocity
         & zvb_adin,     & ! Adjoint input: velocity
         & zrd_adout,    & ! Adjoint output:
         & z3r             ! 3D random field

      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zsshn_tlin,   & ! Tangent input : horizontal gradient
         & zsshb_tlout,  & ! Tangent ouput : horizontal gradient
         & zgtu_tlout,   & ! Tangent output: horizontal gradient
         & zgtv_tlout,   & ! Tangent output: horizontal gradient
         & zgsu_tlout,   & ! Tangent output: horizontal gradient
         & zgsv_tlout,   & ! Tangent output: horizontal gradient
         & zgru_tlout,   & ! Tangent output: horizontal gradient
         & zgrv_tlout,   & ! Tangent output: horizontal gradient
         & zgtu_adin,    & ! Adjoint input : horizontal gradient
         & zgtv_adin,    & ! Adjoint input : horizontal gradient
         & zgsu_adin,    & ! Adjoint input : horizontal gradient
         & zgsv_adin,    & ! Adjoint input : horizontal gradient
         & zgru_adin,    & ! Adjoint input : horizontal gradient
         & zgrv_adin,    & ! Adjoint input : horizontal gradient
         & zsshb_adin,   & ! Adjoint input : horizontal gradient 
         & zsshn_adout,  & ! Adjoint output : horizontal gradient
         & z2r             ! 3D random field

      REAL(KIND=wp) ::   &
                           ! random field standard deviation for:
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp1_1,       & !   scalar product components
         & zsp1_2,       & 
         & zsp1_3,       & !   scalar product components
         & zsp1_4,       &  
         & zsp1_5,       & !   scalar product components
         & zsp1_6,       & 
         & zsp1_7,       & !   scalar product components
         & zsp1_8,       &
         & zsp1_9,       &
         & zsp1_10,      &
         & zsp1_11,      &
         & zsp1_12,      &
         & zsp1_13,      &
         & zsp2,         & ! scalar product involving the adjoint routine
         & zsp2_1,       & !   scalar product components
         & zsp2_2,       & 
         & zsp2_3,       &
         & zsp2_4,       &
         & zsp2_5,       &
         & zsp2_6

      CHARACTER (LEN=14) :: &
         & cl_name

      ! Allocate memory
      ALLOCATE( &
         & ztn_tlin(jpi,jpj,jpk),     & 
         & zsn_tlin(jpi,jpj,jpk),     & 
         & zun_tlin(jpi,jpj,jpk),     & 
         & zvn_tlin(jpi,jpj,jpk),     &
         & zrd_tlin(jpi,jpj,jpk),     &
         & zsshn_tlin(jpi,jpj),       &  
         & ztn_adout(jpi,jpj,jpk),    & 
         & zsn_adout(jpi,jpj,jpk),    &
         & zun_adout(jpi,jpj,jpk),    & 
         & zvn_adout(jpi,jpj,jpk),    &
         & zrd_adout(jpi,jpj,jpk),    &
         & zsshn_adout(jpi, jpj),     & 
         & z3r(jpi,jpj,jpk),          & 
         & z2r(jpi,jpj),              &        
         & zub_tlout(jpi,jpj,jpk),    & 
         & zvb_tlout(jpi,jpj,jpk),    &
         & zsb_tlout(jpi,jpj,jpk),    & 
         & ztb_tlout(jpi,jpj,jpk),    &
         & zsshb_tlout(jpi,jpj),      &
         & zgtu_tlout(jpi,jpj),       & 
         & zgtv_tlout(jpi,jpj),       & 
         & zgsu_tlout(jpi,jpj),       & 
         & zgsv_tlout(jpi,jpj),       & 
         & zgru_tlout(jpi,jpj),       & 
         & zgrv_tlout(jpi,jpj),       &
         & zrd_tlout(jpi,jpj,jpk),    & 
         & zrhop_tlout(jpi,jpj,jpk),  &
         & zub_adin(jpi,jpj,jpk),     & 
         & zvb_adin(jpi,jpj,jpk),     &
         & zsb_adin(jpi,jpj,jpk),     & 
         & ztb_adin(jpi,jpj,jpk),     &
         & zsshb_adin(jpi,jpj),       &
         & zgtu_adin(jpi,jpj),        & 
         & zgtv_adin(jpi,jpj),        & 
         & zgsu_adin(jpi,jpj),        & 
         & zgsv_adin(jpi,jpj),        & 
         & zgru_adin(jpi,jpj),        & 
         & zgrv_adin(jpi,jpj),        &
         & zrd_adin(jpi,jpj,jpk),     &
         & zrhop_adin(jpi,jpj,jpk)    &
         & ) 


      !=============================================================
      ! 1) dx = ( T ) and dy = ( T )
      !=============================================================

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
      ztn_tlin   (:,:,:)  = 0.0_wp     
      zsn_tlin   (:,:,:)  = 0.0_wp     
      zrd_tlin   (:,:,:)  = 0.0_wp     
      ztn_adout  (:,:,:)  = 0.0_wp    
      zsn_adout  (:,:,:)  = 0.0_wp    
      zrd_adout  (:,:,:)  = 0.0_wp    
      zgtu_tlout (:,:  )  = 0.0_wp    
      zgtv_tlout (:,:  )  = 0.0_wp    
      zgsu_tlout (:,:  )  = 0.0_wp    
      zgsv_tlout (:,:  )  = 0.0_wp    
      zgru_tlout (:,:  )  = 0.0_wp    
      zgrv_tlout (:,:  )  = 0.0_wp
      zrd_tlout  (:,:,:) = 0.0_wp
      zrhop_tlout(:,:,:) = 0.0_wp    
      zgtu_adin  (:,:  ) = 0.0_wp     
      zgtv_adin  (:,:  ) = 0.0_wp     
      zgsu_adin  (:,:  ) = 0.0_wp     
      zgsv_adin  (:,:  ) = 0.0_wp     
      zgru_adin  (:,:  ) = 0.0_wp     
      zgrv_adin  (:,:  ) = 0.0_wp
      zrd_adin   (:,:,:) = 0.0_wp
      zrhop_adin (:,:,:) = 0.0_wp

      tn_tl      (:,:,:) = 0.0_wp
      sn_tl      (:,:,:) = 0.0_wp
      un_tl      (:,:,:) = 0.0_wp
      vn_tl      (:,:,:) = 0.0_wp
      sshn_tl    (:,:  ) = 0.0_wp
      tb_tl      (:,:,:) = 0.0_wp
      sb_tl      (:,:,:) = 0.0_wp
      ub_tl      (:,:,:) = 0.0_wp
      vb_tl      (:,:,:) = 0.0_wp
      sshb_tl    (:,:  ) = 0.0_wp
      rhd_tl     (:,:,:) = 0.0_wp
      rhop_tl    (:,:,:) = 0.0_wp
      gtu_tl     (:,:  ) = 0.0_wp
      gsu_tl     (:,:  ) = 0.0_wp
      gru_tl     (:,:  ) = 0.0_wp
      gtv_tl     (:,:  ) = 0.0_wp
      gsv_tl     (:,:  ) = 0.0_wp
      grv_tl     (:,:  ) = 0.0_wp
      tb_ad      (:,:,:) = 0.0_wp
      sb_ad      (:,:,:) = 0.0_wp
      ub_ad      (:,:,:) = 0.0_wp
      vb_ad      (:,:,:) = 0.0_wp
      sshb_ad    (:,:  ) = 0.0_wp
      tn_ad      (:,:,:) = 0.0_wp
      sn_ad      (:,:,:) = 0.0_wp
      un_ad      (:,:,:) = 0.0_wp
      vn_ad      (:,:,:) = 0.0_wp
      sshn_ad    (:,:  ) = 0.0_wp
      gtu_ad     (:,:  ) = 0.0_wp
      gsu_ad     (:,:  ) = 0.0_wp
      gtv_ad     (:,:  ) = 0.0_wp
      gsv_ad     (:,:  ) = 0.0_wp

      ! Warning, following variables used by istate
      hdivn_tl           = 0.0_wp
      hdivb_tl           = 0.0_wp
      rotn_tl            = 0.0_wp
      rotb_tl            = 0.0_wp
      hdivn_ad           = 0.0_wp
      hdivb_ad           = 0.0_wp
      rotn_ad            = 0.0_wp
      rotb_ad            = 0.0_wp

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 284035 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stdt )
      DO jk = 1, jpk
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               ztn_tlin(ji,jj,jk) = z3r(ji,jj,jk)  
            END DO
         END DO
      END DO 
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - (  471426 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stds )
      DO jk = 1, jpk
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zsn_tlin(ji,jj,jk) = z3r(ji,jj,jk)  
            END DO
         END DO
      END DO 
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 395703 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stdr )
      DO jk = 1, jpk
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zrd_tlin(ji,jj,jk) = z3r(ji,jj,jk)  
            END DO
         END DO
      END DO 
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 12672456 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z2r, 'T', 0.0_wp, stdssh )
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zsshn_tlin(ji,jj) = z2r(ji,jj)  
         END DO
      END DO 
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 456953 + &
              &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z3r, 'U', 0.0_wp, stdu )
      DO jk = 1, jpk
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zun_tlin(ji,jj,jk) = z3r(ji,jj,jk)  
            END DO
         END DO
      END DO
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 3434334 + &
              &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z3r, 'V', 0.0_wp, stdv )
      DO jk = 1, jpk
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zvn_tlin(ji,jj,jk) = z3r(ji,jj,jk)  
            END DO
         END DO
      END DO

      tn_tl  (:,:,:) = ztn_tlin  (:,:,:)
      sn_tl  (:,:,:) = zsn_tlin  (:,:,:)
      rhd_tl (:,:,:) = zrd_tlin  (:,:,:)
      sshn_tl(:,:  ) = zsshn_tlin(:,:  )
      un_tl  (:,:,:) = zun_tlin  (:,:,:)
      vn_tl  (:,:,:) = zvn_tlin  (:,:,:)

     !--------------------------------------------------------------------
     ! Call the tangent routine: dy = L dx
     !--------------------------------------------------------------------

     CALL istate_init_tan

     zrd_tlout   (:,:,:) = rhd_tl  (:,:,:)
     zrhop_tlout (:,:,:) = rhop_tl (:,:,:)
     zgtu_tlout  (:,:  ) = gtu_tl  (:,:  )
     zgtv_tlout  (:,:  ) = gtv_tl  (:,:  )
     zgru_tlout  (:,:  ) = gru_tl  (:,:  )
     zgrv_tlout  (:,:  ) = grv_tl  (:,:  )
     zgsu_tlout  (:,:  ) = gsu_tl  (:,:  )
     zgsv_tlout  (:,:  ) = gsv_tl  (:,:  )
     zsshb_tlout (:,:  ) = sshb_tl (:,:  )
     ztb_tlout   (:,:,:) = tb_tl   (:,:,:)
     zsb_tlout   (:,:,:) = sb_tl   (:,:,:)
     zub_tlout   (:,:,:) = ub_tl   (:,:,:)
     zvb_tlout   (:,:,:) = vb_tl   (:,:,:)
     zsshb_tlout (:,:  ) = sshb_tl (:,:  )

     !--------------------------------------------------------------------
     ! Initialize the adjoint variables: dy^* = W dy
     !--------------------------------------------------------------------

     DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zrd_adin(ji,jj,jk)   = zrd_tlout(ji,jj,jk) &
                 &                 * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)&
                 &                 * tmask(ji,jj,jk)
              zrhop_adin(ji,jj,jk) = zrhop_tlout(ji,jj,jk) &
                 &                 * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)&
                 &                 * tmask(ji,jj,jk)
           END DO
        END DO
     END DO
     DO jj = nldj, nlej
        DO ji = nldi, nlei
           zgtu_adin(ji,jj) = zgtu_tlout(ji,jj) &
              &             * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,1) &
              &             * umask(ji,jj,1)
           zgsu_adin(ji,jj) = zgsu_tlout(ji,jj) &
              &             * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,1) &
              &             * umask(ji,jj,1)
           zgru_adin(ji,jj) = zgru_tlout(ji,jj) &
              &             * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,1) &
              &             * umask(ji,jj,1)
           zgtv_adin(ji,jj) = zgtv_tlout(ji,jj) &
              &             * e1v(ji,jj) * e2v(ji,jj) * fse3v(ji,jj,1) &
              &             * vmask(ji,jj,1)
           zgsv_adin(ji,jj) = zgsv_tlout(ji,jj) &
              &             * e1v(ji,jj) * e2v(ji,jj) * fse3v(ji,jj,1) &
              &             * vmask(ji,jj,1)
           zgrv_adin(ji,jj) = zgrv_tlout(ji,jj) &
              &             * e1v(ji,jj) * e2v(ji,jj) * fse3v(ji,jj,1) &
              &             * vmask(ji,jj,1)
        END DO
     END DO
     DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              ztb_adin(ji,jj,jk)   = ztb_tlout(ji,jj,jk) &
                 &                 * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)&
                 &                 * tmask(ji,jj,jk)
           END DO
        END DO
     END DO
     DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zub_adin(ji,jj,jk)   = zub_tlout(ji,jj,jk) &
                 &                 * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,jk)&
                 &                 * umask(ji,jj,jk)
           END DO
        END DO
     END DO
     DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zvb_adin(ji,jj,jk)   = zvb_tlout(ji,jj,jk) &
                 &                 * e1v(ji,jj) * e2v(ji,jj) * fse3v(ji,jj,jk)&
                 &                 * vmask(ji,jj,jk)
           END DO
        END DO
     END DO
     DO jj = nldj, nlej
        DO ji = nldi, nlei
           zsshb_adin(ji,jj)      = zsshb_tlout(ji,jj) &
              &                 * e1t(ji,jj) * e2t(ji,jj) * wesp_ssh &
              &                 * tmask(ji,jj,1)
        END DO
     END DO

      !--------------------------------------------------------------------
      ! Compute the scalar product: ( L dx )^T W dy
      !--------------------------------------------------------------------

      zsp1_1    = DOT_PRODUCT( zrd_tlout    , zrd_adin    )
      zsp1_2    = DOT_PRODUCT( zrhop_tlout  , zrhop_adin  )
      zsp1_3    = DOT_PRODUCT( zgtu_tlout   , zgtu_adin   )
      zsp1_4    = DOT_PRODUCT( zgru_tlout   , zgru_adin   )
      zsp1_5    = DOT_PRODUCT( zgsu_tlout   , zgsu_adin   )
      zsp1_6    = DOT_PRODUCT( zgtv_tlout   , zgtv_adin   )
      zsp1_7    = DOT_PRODUCT( zgrv_tlout   , zgrv_adin   )
      zsp1_8    = DOT_PRODUCT( zgsv_tlout   , zgsv_adin   )
      zsp1_9    = DOT_PRODUCT( zub_tlout    , zub_adin    )
      zsp1_10   = DOT_PRODUCT( zvb_tlout    , zvb_adin    )
      zsp1_11   = DOT_PRODUCT( ztb_tlout    , ztb_adin    )
      zsp1_12   = DOT_PRODUCT( zsb_tlout    , zsb_adin    )
      zsp1_13   = DOT_PRODUCT( zsshb_tlout  , zsshb_adin  )

      zsp1      = zsp1_1 + zsp1_2  + zsp1_3  + zsp1_4  + &
                & zsp1_5 + zsp1_6  + zsp1_7  + zsp1_8  + &
                & zsp1_9 + zsp1_10 + zsp1_11 + zsp1_12 + zsp1_13 

      !--------------------------------------------------------------------
      ! Call the adjoint routine: dx^* = L^T dy^*
      !--------------------------------------------------------------------

      rhd_ad  (:,:,:) = zrd_adin   (:,:,:)
      rhop_ad (:,:,:) = zrhop_adin (:,:,:)
      gtu_ad  (:,:  ) = zgtu_adin  (:,:  )
      gtv_ad  (:,:  ) = zgtv_adin  (:,:  )
      gru_ad  (:,:  ) = zgru_adin  (:,:  )
      grv_ad  (:,:  ) = zgrv_adin  (:,:  )
      gsu_ad  (:,:  ) = zgsu_adin  (:,:  )
      gsv_ad  (:,:  ) = zgsv_adin  (:,:  )
      ub_ad   (:,:,:) = zub_adin   (:,:,:)
      vb_ad   (:,:,:) = zvb_adin   (:,:,:)
      tb_ad   (:,:,:) = ztb_adin   (:,:,:)
      sb_ad   (:,:,:) = zsb_adin   (:,:,:)
      sshb_ad (:,:  ) = zsshb_adin (:,:  ) 
     
      CALL istate_init_adj

      ztn_adout  (:,:,:) = tn_ad   (:,:,:)
      zsn_adout  (:,:,:) = sn_ad   (:,:,:)
      zrd_adout  (:,:,:) = rhd_ad  (:,:,:)
      zun_adout  (:,:,:) = un_ad   (:,:,:)
      zvn_adout  (:,:,:) = vn_ad   (:,:,:)
      zsshn_adout(:,:  ) = sshn_ad (:,:  )

      !--------------------------------------------------------------------
      ! Compute the scalar product: dx^T L^T W dy
      !--------------------------------------------------------------------
       
      zsp2_1    = DOT_PRODUCT( ztn_tlin  , ztn_adout    )
      zsp2_2    = DOT_PRODUCT( zsn_tlin  , zsn_adout    )
      zsp2_3    = DOT_PRODUCT( zrd_tlin  , zrd_adout    )
      zsp2_4    = DOT_PRODUCT( zun_tlin  , zun_adout    )
      zsp2_5    = DOT_PRODUCT( zvn_tlin  , zvn_adout    )
      zsp2_6    = DOT_PRODUCT( zsshn_tlin, zsshn_adout  )

      zsp2      = zsp2_1 + zsp2_2 + zsp2_3 + zsp2_4 + zsp2_5 + zsp2_6

      ! Compare the scalar products

      ! 14 char:'12345678901234'
      cl_name = 'istate_tst    '
      CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )

      ! Deallocate memory
      DEALLOCATE(        &
         & ztn_tlin,     & 
         & zsn_tlin,     & 
         & zrd_tlin,     &  
         & ztn_adout,    & 
         & zsn_adout,    &
         & zrd_adout,    & 
         & z3r,          & 
         & z2r,          &
         & zub_tlout,    & 
         & zvb_tlout,    &
         & zsb_tlout,    & 
         & ztb_tlout,    &
         & zsshb_tlout,  &        
         & zgtu_tlout,   & 
         & zgtv_tlout,   & 
         & zgsu_tlout,   & 
         & zgsv_tlout,   & 
         & zgru_tlout,   & 
         & zgrv_tlout,   &
         & zrd_tlout,    & 
         & zrhop_tlout,  &
         & zub_adin,     & 
         & zvb_adin,     &
         & zsb_adin,     & 
         & ztb_adin,     &
         & zsshb_adin,   &
         & zgtu_adin,    & 
         & zgtv_adin,    & 
         & zgsu_adin,    & 
         & zgsv_adin,    & 
         & zgru_adin,    & 
         & zgrv_adin,    &
         & zrd_adin,     &
         & zrhop_adin    &
         & )

   END SUBROUTINE istate_init_adj_tst

   !!=====================================================================
#endif
END MODULE istate_tam
