MODULE nemotam_hlt
#if defined key_tam
   !!==========================================================================
   !!                       ***  MODULE nemotam ***
   !! NEMOTAM system : Tangent and Adjoint for NEMO.
   !!==========================================================================

   !!----------------------------------------------------------------------
   !!   nemotam_main   : Main driver routine for NEMOTAM
   !!   nemotam_init   : Initialization of NEMOTAM
   !!   nemotam_final  : Finish up NEMOTAM
   !!----------------------------------------------------------------------
   !! History :
   !!   1.0  !  07-06  (K. Mogensen) Initial version
   !!        !  09-06  (F. Vigilant)  Modified to split NEMOVAR / NEMOTAM
   !!                                     module tamctl     i/o varctl
   !!---------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !! * Modules used
   USE hlttst          ! testing routine of linear-tangent hypothesis
   USE par_tlm
   USE in_out_manager  ! I/O manager

   IMPLICIT NONE
   PRIVATE

   !! * Module variables

   !! * Routine accessibility

   PUBLIC &
      & nemotam_hlt_main  

CONTAINS

   SUBROUTINE nemotam_hlt_main 
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE nemotam_main  ***
      !!          
      !! ** Purpose : Main driver routine for NEMOTAM
      !!
      !! ** Method  : Nothing yet
      !!
      !! ** Action  : Nothing yet
      !!
      !! History :
      !!        !  07-06  (K. Mogensen) Original code
      !!----------------------------------------------------------------------

      !! * Local declarations

      ! Read namelist for model  
      CALL hlt_root

      ! Main tst
      CALL hlt_tst

   END SUBROUTINE nemotam_hlt_main

   SUBROUTINE hlt_root
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE nemovar_root ***
      !!          
      !! ** Purpose : Choose which init must be done according to test
      !!
      !! ** Method  : Read namelist
      !!
      !! ** Action  : 
      !!
      !! History :
      !!        !  10-07  (F. Vigilant) Original code
      !!----------------------------------------------------------------------
      !! * Modules used
      USE par_tlm
!      USE in_out_manager, ONLY: & ! I/O manager
!      & numnam, numout
      !! * Arguments

      !! * Local declarations
      NAMELIST/namhlt/ nstg, ln_hltt, ln_hlts, ln_hltuv, ln_hltssh, c_hltinc, &
                     & ln_incdx, ln_hnorm, rhstdt, rhstds, rhstduv, rhstdssh 

      CHARACTER (len=20) ::  namelistname = 'namelist.hlt'
      CHARACTER (len=28) ::  file_out
      !! * Arguments

      nstg = 0
      ln_hltt   = .FALSE.
      ln_hlts   = .FALSE.
      ln_hltuv  = .FALSE.
      ln_hltssh = .FALSE.
      c_hltinc  = ''
      ln_incdx  = .FALSE.
      ln_hnorm  = .FALSE.
      rhstdt    = 1.0_wp
      rhstds    = 0.1_wp
      rhstduv   = 0.01_wp
      rhstdssh  = 0.01_wp

      file_out = 'nemohlt.output'
     
      ! open listing and namelist units
      CALL ctlopn( numout, file_out, 'UNKNOWN', 'FORMATTED',   &
         &         'SEQUENTIAL', 1, 6, .FALSE., 1 )

      WRITE(numout,*)
      WRITE(numout,*) '                 L O D Y C - I P S L'
      WRITE(numout,*) '                     O P A model'
      WRITE(numout,*) '            Ocean General Circulation Model'
      WRITE(numout,*) '               version OPA 9.0  (2005) '
      WRITE(numout,*)
      WRITE(numout,*)
      call flush(numout)

      CALL ctlopn( numnam,namelistname, 'OLD', 'FORMATTED', 'SEQUENTIAL',   &
         &           1, numout, .FALSE., 1 )

      REWIND( numnam )
      READ  ( numnam, namhlt )

   END SUBROUTINE hlt_root
#endif 
END MODULE nemotam_hlt
