MODULE opatam_tst_ini
#if defined key_tam
#if defined key_tst_tlm
   !!==============================================================================
   !!                       ***  MODULE opatam_tst_init   ***
   !! Initialization of OPA and OPATAM for Tangent Linear and Adjoint modules
   !!==============================================================================

   !!----------------------------------------------------------------------
   !!   opa_4_tst_ini    : initialization of the opa model
   !!   opatam_4_tst_ini : initalization of the opatam model 
   !!   opa_hdr_ini      : initialization of the Header Print out file
   !!----------------------------------------------------------------------
   !! History :
   !!   9.0  !  09-06  (F. Vigilant)  Original code
   !!                                 Mainframe from opa.F90 (07-07)
   !!----------------------------------------------------------------------
   !! * Modules used
   USE oce             ! dynamics and tracers variables
   USE cpl_oce         ! ocean-atmosphere-sea ice coupled exchanges
   USE dom_oce         ! ocean space domain variables
   USE sbc_oce         ! surface boundary condition: ocean
   USE trdmod_oce      ! ocean variables trends
   USE daymod          ! calendar
   USE in_out_manager  ! I/O manager
   USE lib_mpp         ! distributed memory computing
   USE domcfg          ! domain configuration               (dom_cfg routine)
   USE mppini          ! shared/distributed memory setting (mpp_init routine)
   USE domain          ! domain initialization             (dom_init routine)
   USE obc_par         ! open boundary cond. parameters
   USE obcini          ! open boundary cond. initialization (obc_ini routine)
   USE bdy_par         ! unstructured open boundary cond. parameters
   USE bdyini          ! unstructured open boundary cond. initialization (bdy_init routine)
   USE istate          ! initial state setting          (istate_init routine)
   USE eosbn2          ! equation of state            (eos bn2 routine)
   USE zpshde          ! partial step: hor. derivative (zps_hde routine)

   ! ocean physics
   USE ldfdyn          ! lateral viscosity setting      (ldfdyn_init routine)
   USE ldftra          ! lateral diffusivity setting    (ldftra_init routine)
   USE zdfini

   USE phycst          ! physical constant                  (par_cst routine)
   USE trdmod          ! momentum/tracers trends       (trd_mod_init routine)

   USE diaptr          ! poleward transports           (dia_ptr_init routine)

#if defined key_oasis3
   USE cpl_oasis3      ! OASIS3 coupling (to ECHAM5)
#elif defined key_oasis4
   USE cpl_oasis4      ! OASIS4 coupling (to ECHAM5)
#endif
   USE dynspg_oce      ! Control choice of surface pressure gradient schemes
   USE prtctl          ! Print control                 (prt_ctl_init routine)
   USE c1d             ! 1D configuration
   USE dyncor_c1d      ! Coriolis factor at T-point
#if defined key_top
   USE trcini          ! Initialization of the passive tracers
#endif
   !! * Modules used for TAM
   USE domcfg          ! domain configuration               (dom_cfg routine)
   USE tamctl          ! Control parameters
   USE tamtrj          ! handling of the trajectory
   USE oce_tam         ! TL and adjoint data
   USE sbc_oce_tam     ! Surface BCs tangent and adjoint arrays
   USE trc_oce_tam     ! Trend tangent and adjoint arrays
   USE sol_oce_tam     ! Solver tangent and adjoint arrays
   ! ocean physics
   USE zdfini 
   USE opa
   USE par_tlm

   IMPLICIT NONE
   PRIVATE

   !! * Module variables
   CHARACTER (len=64) ::        &
      cform_aaa="( /, 'AAAAAAAA', / ) "     ! flag for output listing
   LOGICAL :: ldirinit = .FALSE.            ! flag if initialization of direct is done

   !! * Routine accessibility
   PUBLIC &
     & opa_opatam_ini,          &
     & opa_4_tst_ini,           &
     & opatam_4_tst_ini,        &
     & tlm_namrd


CONTAINS 

   SUBROUTINE opa_opatam_ini
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE opa_4_tst_ini  ***
      !!
      !! ** Purpose :   initialization of the opa+opatam model
      !!
      !! History :
      !!   9.0  !  09-07  (F. Vigilant)  
      !!                                 
      !!----------------------------------------------------------------------

      ! Initialization
      CALL opa_hdr_ini
      IF(lwp) WRITE(numout,cform_aaa)       ! Flag AAAAAAA
      CALL opa_flg                          ! Control prints & Benchmark
      CALL tlm_namrd

      SELECT CASE (tlm_bch)
         CASE ( 0, 1 )
      CALL  opa_4_tst_ini
         CASE ( 2 )
      CALL  opatam_4_tst_ini
         CASE DEFAULT
            CALL ctl_stop( '        Wrong Value of tlm_bch')
      END SELECT


   END SUBROUTINE opa_opatam_ini

   SUBROUTINE opa_4_tst_ini
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE opa_4_tst_ini  ***
      !!
      !! ** Purpose :   initialization of the opa model
      !!
      !! History :
      !!   9.0  !  09-06  (F. Vigilant)  import from opa_ini
      !!                                 
      !!----------------------------------------------------------------------
 
#if defined key_coupled
      INTEGER ::   itro, istp0        ! ???
#endif

      ! Initializations
      ! ===============

      !                                     ! ============================== !
      !                                     !  Model general initialization  !
      !                                     ! ============================== !

                                            ! Domain decomposition
      IF( jpni*jpnj == jpnij ) THEN
         CALL mpp_init                          ! standard cutting out
      ELSE
         CALL mpp_init2                         ! eliminate land processors
      ENDIF
      
      CALL phy_cst                          ! Physical constants

      CALL dom_cfg                          ! Domain configuration
      
      CALL dom_init                         ! Domain

      IF( lk_c1d    ) THEN                      ! adaptation for 1D configuration 
         CALL cor_c1d                                ! redefine Coriolis at T-point
         umask(:,:,:) = tmask(:,:,:)                 ! U, V and T-points are the same
         vmask(:,:,:) = tmask(:,:,:)                 ! 
      ENDIF

      IF( ln_ctl    )   CALL prt_ctl_init   ! Print control

      IF( lk_obc    )   CALL obc_init       ! Open boundaries 

      IF( lk_bdy    )   CALL bdy_init       ! Unstructured open boundaries

!      CALL istate_init                      ! ocean initial state (Dynamics and tracers)
!      CALL istate_p
      !                                     ! Ocean physics

      CALL ldf_dyn_init                         ! Lateral ocean momentum physics

      CALL ldf_tra_init                         ! Lateral ocean tracer physics

      CALL zdf_init                             ! Vertical ocean physics

      CALL trd_mod_init                         ! Mixed-layer/Vorticity/Integral constraints trends

#if defined key_top
      CALL trc_ini                           ! Passive tracers
#endif

      IF(lwp) WRITE(numout,*)'Euler time step switch is ', neuler
      ldirinit = .TRUE.

      CALL     tam_trj_ini

      CALL day_init
      CALL day(nit000)

      IF(lwp) WRITE(numout,cform_aaa)       ! Flag AAAAAAA

   END SUBROUTINE opa_4_tst_ini

   SUBROUTINE opa_hdr_ini
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE opa_hdr_ini  ***
      !!
      !! ** Purpose :   initialization of the Header Print out file
      !!
      !! History :
      !!   9.0  !  09-06  (F. Vigilant)  extract from opa_ini
      !!                                 
      !!----------------------------------------------------------------------

      ! local declaration
#if defined key_oasis3 || defined key_oasis4
      INTEGER :: localComm
#endif
      CHARACTER (len=20) ::   namelistname
      CHARACTER (len=28) ::   file_out
      NAMELIST/namctl/ ln_ctl, nprint, nictls, nictle,   &
         &             isplt , jsplt , njctls, njctle, nbench, nbit_cmp
      ! Initializations
      ! ===============

      file_out = 'nemotam_tst.output'
      
      ! open listing and namelist units
      CALL ctlopn( numout, file_out, 'UNKNOWN', 'FORMATTED',   &
         &         'SEQUENTIAL', 1, 6, .FALSE., 1 )

      WRITE(numout,*)
      WRITE(numout,*) '                 L O D Y C - I P S L'
      WRITE(numout,*) '                     O P A model'
      WRITE(numout,*) '            Ocean General Circulation Model'
      WRITE(numout,*) '               version OPA 9.0  (2005) '
      WRITE(numout,*)
      WRITE(numout,*)

      ! Namelist namctl : Control prints & Benchmark
      REWIND( numnam )
      READ  ( numnam, namctl )

#if defined key_oasis3 || defined key_oasis4
      call cpl_prism_init(localComm)
      ! Nodes selection
      narea = mynode(localComm)
#else
      ! Nodes selection
      nproc = mynode()
      narea = nproc + 1    ! mynode return the rank of proc (0 --> jpnij -1 )
#endif
      narea = narea + 1    ! mynode return the rank of proc (0 --> jpnij -1 )
      lwp   = narea == 1

      ! open additionnal listing
!      IF( ln_ctl )   THEN
         IF( narea-1 > 0 )   THEN
            WRITE(file_out,FMT="('nemotam_tst.output_',I4.4)") narea-1
            CALL ctlopn( numout, file_out, 'UNKNOWN', 'FORMATTED',   &
               &         'SEQUENTIAL', 1, numout, .FALSE., 1 )
            lwp = .TRUE.
         ENDIF
            !
!      ENDIF
      IF(lwp) THEN
            WRITE(numout,*)
            WRITE(numout,*) '                 L O D Y C - I P S L'
            WRITE(numout,*) '                     O P A model'
            WRITE(numout,*) '            Ocean General Circulation Model'
            WRITE(numout,*) '               version OPA 9.0  (2005) '
            WRITE(numout,*) '                   MPI Ocean output '
            WRITE(numout,*)
            WRITE(numout,*)
      ENDIF


   END SUBROUTINE opa_hdr_ini

   SUBROUTINE opatam_4_tst_ini
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE nemotam_init  ***
      !!          
      !! ** Purpose : Initialize grids
      !!
      !! ** Method  : Read the namelist and call reading routines
      !!
      !! ** Action  : Read the namelist and call reading routines
      !!
      !! History :
      !!        !  09-06  (F. Vigilant) Extract from nemovar_init routine
      !!----------------------------------------------------------------------

      !! * Local declarations
      CHARACTER (len=128) :: file_out !!= 'nemovar.output'


      ! Nodes selection
      nproc = mynode()
      narea = nproc + 1    ! mynode return the rank of proc (0 --> jpnij -1 )
      lwp   = narea == 1

      ! open additionnal listing
      IF( narea-1 > 0 )   THEN
!!         IF ( .NOT. ldirinit) THEN
!!            WRITE(file_out,FMT="('nemovar.output_',I4.4)") narea-1
!!            IF( numout /= 0 .AND. numout /= 6 ) THEN 
!!               CALL ctlopn( numout, file_out, 'UNKNOWN', 'FORMATTED',   &
!!               &         'SEQUENTIAL', 1, numout, .FALSE., 1 )
!!            ENDIF
!         CALL nemotam_banner( numout )
!!         ENDIF
         lwp = .TRUE.
         !
      ENDIF      

      IF(lwp) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'nemotam_init: Control prints & Benchmark'
         WRITE(numout,*) '~~~~~~~~~~~~ '
         WRITE(numout,*) '          Namelist namctl'
         WRITE(numout,*) '             run control (for debugging)     ln_ctl    = ', ln_ctl
         WRITE(numout,*) '             level of print                  nprint    = ', nprint
         WRITE(numout,*) '             bit comparison mode (0/1)       nbit_cmp  = ', nbit_cmp
      ENDIF

      IF ( .NOT. ldirinit ) THEN
        IF( jpni*jpnj == jpnij ) THEN
           CALL mpp_init                      ! standard cutting out
        ELSE
           CALL mpp_init2                     ! eliminate land processors
        ENDIF
      ENDIF

      IF (lwp) THEN
         ! Diagnostic file for tangent test
         WRITE(file_out,FMT="('tan_diag.output_',I4.4)") , narea-1
         CALL ctlopn( numtan, file_out, 'UNKNOWN', 'FORMATTED',   &
            &         'SEQUENTIAL', 1, numtan, .FALSE., 1 )

         WRITE(numtan,*) 'Routine loop      p      Nn=M(X+dX)-M(X)         En=Nn / L(hdX)       ', &
              & 'Er = (Nn -L)/L            L                  Nn-L                   (En-1)/p   ', &
              & '         Er/p'
         WRITE(numtan,*) ' ---------------------------------------------------------------------',&
              & '-------------------------------------------------------------------------------',&
              & '-------------'
      ENDIF

      IF (lwp) THEN
         ! Diagnostic file for tangent test (scalar sampling)
         WRITE(file_out,FMT="('tan_diag_sc.output_',I4.4)") , narea-1
         CALL ctlopn( numtan_sc, file_out, 'UNKNOWN', 'FORMATTED',   &
            &         'SEQUENTIAL', 1, numtan_sc, .FALSE., 1 )

         WRITE(numtan_sc,*) 'Routine  scalar    loop    p     index i_pos. ', &
              & 'j_pos.       scalar value'
         WRITE(numtan_sc,*) ' ---------------------------------------------',&
              & '-------------------------'
      ENDIF

      IF ( .NOT. ldirinit ) THEN
         CALL phy_cst                          ! Physical constants
         CALL dom_cfg                          ! Domain configuration  
         CALL dom_init                         ! Domain

          ! Ocean physics
#if defined key_tam
         CALL ldf_dyn_init                         ! Lateral ocean momentum physics
         CALL ldf_tra_init                         ! Lateral ocean tracer physics
#endif
         CALL zdf_init                             ! Vertical ocean physics
      ENDIF
       
      CALL     oce_tam_init( 1 )            ! OCE TAM field
      IF ( ldirinit ) THEN
         CALL sol_oce_tam_init( 1 )            ! Initialize solver tangent variables
      ELSE
         CALL sol_oce_tam_init( 0 )            ! Initialize solver tangent variables
      ENDIF
      CALL trc_oce_tam_init( 1 )            ! TRC TAM fields
#if defined key_tam
      CALL sbc_oce_tam_init( 1 )            ! SBC TAM fields
#endif

      IF ( .NOT. ldirinit) CALL     tam_trj_ini

      CALL day_init
      IF ( .NOT. ldirinit) CALL day( nit000 )

#ifdef EVENTUALLY

#endif
      neuler=1

      IF(lwp) WRITE(numout,cform_aaa)       ! Flag AAAAAAA

   END SUBROUTINE opatam_4_tst_ini

   SUBROUTINE opa_flg
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE opa  ***
      !!
      !! ** Purpose :   Initialize logical flags that control the choice of
      !!      some algorithm or control print
      !!
      !! ** Method  :    Read in namilist namflg logical flags
      !!
      !! History :
      !!   9.0  !  03-11  (G. Madec)  Original code
      !!----------------------------------------------------------------------
      !! * Local declarations

      NAMELIST/namflg/ ln_dynhpg_imp, nn_dynhpg_rst
      !!----------------------------------------------------------------------

      ! Parameter control and print
      ! ---------------------------
      IF(lwp) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'opa_flg: Control prints & Benchmark'
         WRITE(numout,*) '~~~~~~~ '
         WRITE(numout,*) '          Namelist namctl'
         WRITE(numout,*) '             run control (for debugging)     ln_ctl    = ', ln_ctl
         WRITE(numout,*) '             level of print                  nprint    = ', nprint
         WRITE(numout,*) '             Start i indice for SUM control  nictls    = ', nictls
         WRITE(numout,*) '             End i indice for SUM control    nictle    = ', nictle
         WRITE(numout,*) '             Start j indice for SUM control  njctls    = ', njctls
         WRITE(numout,*) '             End j indice for SUM control    njctle    = ', njctle
         WRITE(numout,*) '             number of proc. following i     isplt     = ', isplt
         WRITE(numout,*) '             number of proc. following j     jsplt     = ', jsplt
         WRITE(numout,*) '             benchmark parameter (0/1)       nbench    = ', nbench
         WRITE(numout,*) '             bit comparison mode (0/1)       nbit_cmp  = ', nbit_cmp
      ENDIF

      ! ... Control the sub-domain area indices for the control prints
      IF( ln_ctl )   THEN
         IF( lk_mpp )   THEN
            ! the domain is forced to the real splitted domain in MPI
            isplt = jpni ; jsplt = jpnj ; ijsplt = jpni*jpnj
         ELSE
            IF( isplt == 1 .AND. jsplt == 1  ) THEN
               CALL ctl_warn( '          - isplt & jsplt are equal to 1',   &
                    &         '          - the print control will be done over the whole domain' )
            ENDIF

            ! compute the total number of processors ijsplt
            ijsplt = isplt*jsplt
         ENDIF

         IF(lwp) WRITE(numout,*)'          - The total number of processors over which the'
         IF(lwp) WRITE(numout,*)'            print control will be done is ijsplt : ', ijsplt

         ! Control the indices used for the SUM control
         IF( nictls+nictle+njctls+njctle == 0 )   THEN
            ! the print control is done over the default area
            lsp_area = .FALSE.
         ELSE
            ! the print control is done over a specific  area
            lsp_area = .TRUE.
            IF( nictls < 1 .OR. nictls > jpiglo )   THEN
               CALL ctl_warn( '          - nictls must be 1<=nictls>=jpiglo, it is forced to 1' )
               nictls = 1
            ENDIF

            IF( nictle < 1 .OR. nictle > jpiglo )   THEN
               CALL ctl_warn( '          - nictle must be 1<=nictle>=jpiglo, it is forced to jpiglo' )
               nictle = jpiglo
            ENDIF

            IF( njctls < 1 .OR. njctls > jpjglo )   THEN
               CALL ctl_warn( '          - njctls must be 1<=njctls>=jpjglo, it is forced to 1' )
               njctls = 1
            ENDIF

            IF( njctle < 1 .OR. njctle > jpjglo )   THEN
               CALL ctl_warn( '          - njctle must be 1<=njctle>=jpjglo, it is forced to jpjglo' )
               njctle = jpjglo
            ENDIF

         ENDIF          ! IF( nictls+nictle+njctls+njctle == 0 )
       ENDIF            ! IF(ln_ctl)

      IF( nbench == 1 )   THEN
         SELECT CASE ( cp_cfg )
         CASE ( 'gyre' )
            CALL ctl_warn( '          The Benchmark is activated ' )
         CASE DEFAULT
            CALL ctl_stop( '          The Benchmark is based on the GYRE configuration: key_gyre must &
               &                      be used or set nbench = 0' )
         END SELECT
      ENDIF

      IF( nbit_cmp == 1 )   THEN
         CALL ctl_warn( '          Bit comparison enabled. Single and multiple processor results must bit compare', &
              &         '          WARNING: RESULTS ARE NOT PHYSICAL.' )
      ENDIF


      ! Read Namelist namflg : algorithm FLaG
      ! --------------------
      REWIND ( numnam )
      READ   ( numnam, namflg )

      ! Parameter control and print
      ! ---------------------------
      IF(lwp) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'opa_flg : Hydrostatic pressure gradient algorithm'
         WRITE(numout,*) '~~~~~~~'
         WRITE(numout,*) '          Namelist namflg : set algorithm flags'
         WRITE(numout,*) '             centered (F) or semi-implicit (T)   ln_dynhpg_imp = ', ln_dynhpg_imp
         WRITE(numout,*) '             hydrostatic pressure gradient'
         WRITE(numout,*) '             add dynhpg implicit variable        nn_dynhpg_rst = ', nn_dynhpg_rst
         WRITE(numout,*) '             in restart ot not nn_dynhpg_rst'
      ENDIF
      IF( .NOT. ln_dynhpg_imp )   nn_dynhpg_rst = 0      ! force no adding dynhpg implicit variables in restart

   END SUBROUTINE opa_flg


   SUBROUTINE opa_closefile
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE opa_closefile  ***
      !!
      !! ** Purpose :   Close the files
      !!
      !! ** Method  :
      !!
      !! History :
      !!   9.0  !  05-01  (O. Le Galloudec)  Original code
      !!----------------------------------------------------------------------
      !! * Modules used
      USE dtatem        ! temperature data
      USE dtasal        ! salinity data
      !!----------------------------------------------------------------------

      IF ( lk_mpp ) CALL mppsync

      ! 1. Unit close
      ! -------------

      CLOSE( numnam )           ! namelist
      CLOSE( numout )           ! standard model output file

      IF(lwp) CLOSE( numstp )   ! time-step file
      IF(lwp) CLOSE( numsol )

   END SUBROUTINE opa_closefile

   SUBROUTINE tlm_namrd
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE tlm_namrd  ***
      !!
      !! ** Purpose :   Control feature of stp_tlm_tst loops
      !!
      !! ** Method  :    Read in namilist namtst_tlm parameters
      !!
      !! History :
      !!   9.0  !  09-07  (F. Vigilant)  Original code
      !!----------------------------------------------------------------------
      !! * Module used
      USE in_out_manager, ONLY: & ! I/O manager
      & numout,                 &
      & numnam,                 &
      & lwp
      !! * Local declarations

      NAMELIST/namtst_tlm/ tlm_bch, cur_loop, h_ratio

      ! Read Namelist namflg : algorithm FLaG
      ! --------------------
      REWIND ( numnam )
      READ   ( numnam, namtst_tlm )

      ! Parameter control and print
      ! ---------------------------
      IF(lwp) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'tlm_namrd : Current loop iteration and h_ratio values '
         WRITE(numout,*) '~~~~~~~~~'
         WRITE(numout,*) '          Namelist namtst_tlm : set algorithm parameters'
         WRITE(numout,*) '             current branch test     = ' , tlm_bch
         WRITE(numout,*) '             current loop iteration  = ' , cur_loop
         WRITE(numout,*) '             current h_ratio applied = ' , h_ratio
         WRITE(numout,*) 
      ENDIF

   END SUBROUTINE tlm_namrd

   !!======================================================================
#endif
#endif
END MODULE opatam_tst_ini
