MODULE par_tlm
   !!======================================================================
   !!                    ***  MODULE  par_tlm  ***
   !! Tangent Model Accuracy:  parameters defined in memory 
   !!=====================================================================
   !!
   !! ** Purpose :   Define in memory Tangent Model accuracy parameters
   !!
   !! History :
   !!   9.0  !  07-07  (F. Vigilant)  Original code
   !!----------------------------------------------------------------------

   !! * Modules used
   USE par_oce, ONLY:       &  !: ocean parameters
     & wp

   IMPLICIT NONE
   PRIVATE

   INTEGER , PUBLIC ::      & !!: namtst_tlm tangent test parameters
      tlm_bch,              &  !: branch value (0:M(x); 1:M(x+dx); 2:LT(dx))
      cur_loop = 1             !: = 1... nmax_loop (=1 means perform only direct 
                               !                       integration)
     
   REAL(wp), PUBLIC ::      & !!: namtst_tlm tangent test parameters
      h_ratio=  1.0_wp         !: ratio applied to the perturbation dX0

   !! namhlt:  assimilation test parameters

   INTEGER , PUBLIC ::      & !!: namhlt test parameters for tangent-linear hypothesis 
      nstg                     !: nstg = 0 running M(X)

   LOGICAL, PUBLIC  ::      & ! namhlt test parameters for tangent-linear hypothesis 
      & ln_hltt,            & ! Switch for temperature increment
      & ln_hlts,            & ! Switch for salinity increment
      & ln_hltuv,           & ! Switch for velocity increment
      & ln_hltssh,          & ! Switch for sea surface height increment
      & ln_hnorm,	    & ! Swith to apply a normalization on increments
      & ln_incdx	      ! dx read from file (T) or compute from restart.nc and restart2.nc (F)
                              ! if (T) filename is increments_dx.nc

   REAL(wp), PUBLIC ::      & ! namhlt test parameters for tangent-linear hypothesis 
      & rhstdt  =  1.0_wp,  & ! temperature normalization factor
      & rhstds  =  0.1_wp,  & ! salinity normalization factor
      & rhstduv =  0.01_wp, & ! velocity normalization factor
      & rhstdssh=  0.01_wp    ! sea surface height normalization factor

   CHARACTER (LEN=100), PUBLIC :: &
      & c_hltinc,           & ! file name of increment
      & c_hltrst1,          & ! file name of restart 1
      & c_hltrst2             ! file name of restart 2 

END MODULE par_tlm
