MODULE step_cpd
   !!======================================================================
   !!                       ***  MODULE step  ***
   !! Time-stepping    : manager of the ocean, tracer and ice time stepping
   !!======================================================================
   !! History :        !  91-03  (G. Madec)  Original code
   !!                  !  92-06  (M. Imbard)  add a first output record
   !!                  !  96-04  (G. Madec)  introduction of dynspg
   !!                  !  96-04  (M.A. Foujols)  introduction of passive tracer
   !!             8.0  !  97-06  (G. Madec)  new architecture of call
   !!             8.2  !  97-06  (G. Madec, M. Imbard, G. Roullet)  free surface
   !!             8.2  !  99-02  (G. Madec, N. Grima)  hpg implicit
   !!             8.2  !  00-07  (J-M Molines, M. Imbard)  Open Bondary Conditions
   !!             9.0  !  02-06  (G. Madec)  free form, suppress macro-tasking
   !!             " "  !  04-08  (C. Talandier) New trends organization
   !!             " "  !  05-01  (C. Ethe) Add the KPP closure scheme
   !!             " "  !  05-11  (V. Garnier) Surface pressure gradient organization
   !!             " "  !  05-11  (G. Madec)  Reorganisation of tra and dyn calls
   !!             " "  !  06-01  (L. Debreu, C. Mazauric)  Agrif implementation
   !!             " "  !  06-07  (S. Masson)  restart using iom
   !!             " "  !  06-08  (G. Madec)  surface module 
   !!    "        " "  !  07-04  (K. Mogensen, A. Weaver, M. Martin) Assimilation interface
   !!             " "  !  07-07  (J. Chanut, A. Sellar) Unstructured open boundaries (BDY)
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   stp            : OPA system time-stepping
   !!----------------------------------------------------------------------
   USE oce             ! ocean dynamics and tracers variables
   USE dom_oce         ! ocean space and time domain variables 
   USE zdf_oce         ! ocean vertical physics variables
   USE ldftra_oce      ! ocean tracer   - trends
   USE ldfdyn_oce      ! ocean dynamics - trends
   USE cpl_oce         ! coupled ocean-atmosphere variables
   USE in_out_manager  ! I/O manager
   USE iom             !
   USE lbclnk

   USE daymod          ! calendar                         (day     routine)

   USE dtatem          ! ocean temperature data           (dta_tem routine)
   USE dtasal          ! ocean salinity    data           (dta_sal routine)
   USE sbcmod          ! surface boundary condition       (sbc     routine)
   USE sbcrnf          ! surface boundary condition: runoff variables

   USE trcstp          ! passive tracer time-stepping      (trc_stp routine)

   USE traqsr          ! solar radiation penetration      (tra_qsr routine)
   USE trasbc          ! surface boundary condition       (tra_sbc routine)
   USE trabbc          ! bottom boundary condition        (tra_bbc routine)
   USE trabbl          ! bottom boundary layer            (tra_bbl routine)
   USE tradmp          ! internal damping                 (tra_dmp routine)
   USE traadv          ! advection scheme control     (tra_adv_ctl routine)
   USE traldf          ! lateral mixing                   (tra_ldf routine)
   USE cla             ! cross land advection             (tra_cla routine)
   !   zdfkpp          ! KPP non-local tracer fluxes      (tra_kpp routine)
   USE trazdf          ! vertical mixing                  (tra_zdf routine)
   USE tranxt          ! time-stepping                    (tra_nxt routine)
   USE tranpc          ! non-penetrative convection       (tra_npc routine)
   USE trachk          ! check values of salinity         (tra_chk routine)

   USE eosbn2          ! equation of state                (eos_bn2 routine)

   USE dynadv          ! advection                        (dyn_adv routine)
   USE dynvor          ! vorticity term                   (dyn_vor routine)
   USE dynhpg          ! hydrostatic pressure grad.       (dyn_hpg routine)
   USE dynldf          ! lateral momentum diffusion       (dyn_ldf routine)
   USE dynzdf          ! vertical diffusion               (dyn_zdf routine)
   USE dynspg_oce      ! surface pressure gradient        (dyn_spg routine)
   USE dynspg          ! surface pressure gradient        (dyn_spg routine)
   USE dynnxt          ! time-stepping                    (dyn_nxt routine)

   USE obc_par         ! open boundary condition variables
   USE obcdta          ! open boundary condition data     (obc_dta routine)
   USE obcrst          ! open boundary cond. restart      (obc_rst routine)
   USE obcrad          ! open boundary cond. radiation    (obc_rad routine)
   USE obcspg          ! open boundary cond  spg          (obc_spg routine)

   USE bdy_par         ! unstructured open boundary data variables
   USE bdydta          ! unstructured open boundary data  (bdy_dta routine)

   USE divcur          ! hor. divergence and curl      (div & cur routines)
   USE cla_div         ! cross land: hor. divergence      (div_cla routine)
   USE wzvmod          ! vertical velocity                (wzv     routine)

   USE ldfslp          ! iso-neutral slopes               (ldf_slp routine)
   USE ldfeiv          ! eddy induced velocity coef.      (ldf_eiv routine)

   USE zdfbfr          ! bottom friction                  (zdf_bfr routine)
   USE zdftke          ! TKE vertical mixing              (zdf_tke routine)
   USE zdfkpp          ! KPP vertical mixing              (zdf_kpp routine)
   USE zdfddm          ! double diffusion mixing          (zdf_ddm routine)
   USE zdfevd          ! enhanced vertical diffusion      (zdf_evd routine)
   USE zdfric          ! Richardson vertical mixing       (zdf_ric routine)
   USE zdfmxl          ! Mixed-layer depth                (zdf_mxl routine)

   USE zpshde          ! partial step: hor. derivative     (zps_hde routine)
   USE ice_oce         ! sea-ice variable

   USE diawri          ! Standard run outputs             (dia_wri routine)
   USE trdicp          ! Ocean momentum/tracers trends    (trd_wri routine)
   USE trdmld          ! mixed-layer trends               (trd_mld routine)
   USE trdmld_rst      ! restart for mixed-layer trends
   USE trdmod_oce      ! ocean momentum/tracers trends
   USE trdvor          ! vorticity budget                 (trd_vor routine)
   USE diagap          ! hor. mean model-data gap         (dia_gap routine)
   USE diahdy          ! dynamic height                   (dia_hdy routine)
   USE diaptr          ! poleward transports              (dia_ptr routine)
   USE diahth          ! thermocline depth                (dia_hth routine)
   USE diafwb          ! freshwater budget                (dia_fwb routine)
   USE diaspr          ! suface pressure (rigid-lid)      (dia_spr routine)
   USE diaobs          ! obs-minus-model (assimilation)   (dia_obs routine)
   USE flo_oce         ! floats variables
   USE floats          ! floats computation               (flo_stp routine)
   USE asminc          ! assimilation increments          
   USE asmbkg          ! writing out background term    
   USE tamtrj          ! writing out state trajectory       
   USE stpctl          ! time stepping control            (stp_ctl routine)
   USE restart         ! ocean restart                    (rst_wri routine)
   USE prtctl          ! Print control                    (prt_ctl routine)
   USE domvvl          ! variable volume                  (dom_vvl routine)

#if defined key_agrif
   USE agrif_opa_sponge ! Momemtum and tracers sponges
#endif

   IMPLICIT NONE
   PRIVATE

   PUBLIC stp_cpd            ! called by opa.F90

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "zdfddm_substitute.h90"
   !!----------------------------------------------------------------------
   !!   OPA 9.0 , LOCEAN-IPSL (2005) 
   !! $Id$
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt) 
   !!----------------------------------------------------------------------

CONTAINS

#if defined key_agrif
   SUBROUTINE stp_cpd( )
#else
   SUBROUTINE stp_cpd( kstp )
#endif
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE stp  ***
      !!                      
      !! ** Purpose : - Time stepping of OPA (momentum and active tracer eqs.)
      !!              - Time stepping of LIM (dynamic and thermodynamic eqs.)
      !!              - Tme stepping  of TRC (passive tracer eqs.)
      !! 
      !! ** Method  : -1- Update forcings and data  
      !!              -2- Update ocean physics 
      !!              -3- Compute the t and s trends 
      !!              -4- Update t and s 
      !!              -5- Compute the momentum trends
      !!              -6- Update the horizontal velocity
      !!              -7- Compute the diagnostics variables (rd,N2, div,cur,w)
      !!              -8- Outputs and diagnostics
      !!----------------------------------------------------------------------
      !! * Arguments
#if defined key_agrif   
      INTEGER             ::   kstp   ! ocean time-step index
#else
      INTEGER, INTENT(in) ::   kstp   ! ocean time-step index
#endif      
      REAL(wp), DIMENSION(jpi,jpj,jpk) :: zta_tmp, zsa_tmp
      INTEGER ::   jk       ! dummy loop indice
      INTEGER ::   indic    ! error indicator if < 0
      !! ---------------------------------------------------------------------

#if defined key_agrif
      kstp = nit000 + Agrif_Nb_Step()
!      IF ( Agrif_Root() .and. lwp) Write(*,*) '---'
!      IF (lwp) Write(*,*) 'Grid Number',Agrif_Fixed(),' time step ',kstp
#endif   
      indic = 1                    ! reset to no error condition

      CALL day( kstp )             ! Calendar

      CALL rst_opn( kstp )         ! Open the restart file

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Update data, open boundaries, surface boundary condition (including sea-ice)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

      IF( lk_dtatem  )   CALL dta_tem( kstp )         ! update 3D temperature data
      IF( lk_dtasal  )   CALL dta_sal( kstp )         ! update 3D salinity data

                         CALL sbc    ( kstp )         ! Sea Boundary Condition (including sea-ice)

      IF( lk_obc     )   CALL obc_dta( kstp )         ! update dynamic and tracer data at open boundaries
      IF( lk_obc     )   CALL obc_rad( kstp )         ! compute phase velocities at open boundaries

      IF( lk_bdy     )   CALL bdy_dta( kstp )         ! update dynamic and tracer data at unstructured open boundary

      IF( ninist == 1 ) THEN                       ! Output the initial state and forcings
                        CALL dia_wri_state( 'output.init' )
                        ninist = 0
      ENDIF


      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Ocean physics update
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      !-----------------------------------------------------------------------
      !  VERTICAL PHYSICS
      !-----------------------------------------------------------------------
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------

                        CALL bn2( tb, sb, rn2 )              ! before Brunt-Vaisala frequency
      
      !                                                     ! Vertical eddy viscosity and diffusivity coefficients
      IF( lk_zdfric )   CALL zdf_ric( kstp )                       ! Richardson number dependent Kz

      IF( lk_zdftke )   CALL zdf_tke( kstp )                       ! TKE closure scheme for Kz
      IF( lk_zdfkpp )   CALL zdf_kpp( kstp )                       ! KPP closure scheme for Kz

      IF( lk_zdfcst )   THEN                                       ! Constant Kz (reset avt, avm[uv] to the background value)
         avt (:,:,:) = avt0 * tmask(:,:,:)
         avmu(:,:,:) = avm0 * umask(:,:,:)
         avmv(:,:,:) = avm0 * vmask(:,:,:)
      ENDIF

      IF( ln_rnf_mouth ) THEN                                ! increase diffusivity at rivers mouths
         DO jk = 2, nkrnf   ;   avt(:,:,jk) = avt(:,:,jk) + 2.e0 * rn_avt_rnf * rnfmsk(:,:)   ;   END DO
      ENDIF

      IF( ln_zdfevd )   CALL zdf_evd( kstp )                 ! enhanced vertical eddy diffusivity

      IF( lk_zdfddm .AND. .NOT. lk_zdfkpp )   &
         &              CALL zdf_ddm( kstp )                 ! double diffusive mixing

                        CALL zdf_bfr( kstp )                 ! bottom friction

                        CALL zdf_mxl( kstp )                 ! mixed layer depth


      !-----------------------------------------------------------------------
      !  LATERAL PHYSICS
      !-----------------------------------------------------------------------
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------
      IF( lk_ldfslp     )   CALL ldf_slp( kstp, rhd, rn2 )       ! before slope of the lateral mixing
#if defined key_traldf_c2d
      IF( lk_traldf_eiv )   CALL ldf_eiv( kstp )                 ! eddy induced velocity coefficient
#endif

#if defined key_top
      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Passive Tracer Model
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------
                             CALL trc_stp( kstp, indic )            ! time-stepping
#endif

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Active tracers
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va arrays are used as workspace in this section
      !-----------------------------------------------------------------------
                             ta(:,:,:) = 0.e0               ! set tracer trends to zero
                             sa(:,:,:) = 0.e0

      IF(  ln_asmiau .AND. &
         & ln_trainc     )   CALL tra_asm_inc( kstp )       ! apply tracer assimilation increment
                             CALL tra_sbc    ( kstp )       ! surface boundary condition
      IF( ln_traqsr      )   CALL tra_qsr    ( kstp )       ! penetrative solar radiation qsr
      IF( lk_trabbc      )   CALL tra_bbc    ( kstp )       ! bottom heat flux
      IF( lk_trabbl_dif  )   CALL tra_bbl_dif( kstp )       ! diffusive bottom boundary layer scheme
      IF( lk_trabbl_adv  )   CALL tra_bbl_adv( kstp )       ! advective (and/or diffusive) bottom boundary layer scheme
      IF( lk_tradmp      )   CALL tra_dmp    ( kstp )       ! internal damping trends
!                             CALL tra_adv    ( kstp )       ! horizontal & vertical advection
!      IF( n_cla == 1     )   CALL tra_cla    ( kstp )       ! Cross Land Advection (Update Hor. advection)
      IF( lk_zdfkpp )        CALL tra_kpp    ( kstp )       ! KPP non-local tracer fluxes
!                             CALL tra_ldf    ( kstp )       ! lateral mixing
#if defined key_agrif
      IF(.NOT. Agrif_Root()) CALL Agrif_Sponge_tra          ! tracers sponge
#endif
!                             CALL tra_zdf    ( kstp )       ! vertical mixing and after tracer fields
      IF( ln_zdfnpc      )   CALL tra_npc    ( kstp )       ! update after fields by non-penetrative convection
!                             CALL tra_nxt    ( kstp )       ! tracer fields at next time step
      IF( ln_dynhpg_imp  ) THEN                             ! semi-implicit hpg
                               CALL eos( ta, sa, rhd, rhop )          ! Time-filtered in situ density used in dynhpg module
         IF( ln_zps    )       CALL zps_hde( kstp, ta, sa, rhd,   &   ! Partial steps: time filtered hor. gradient
            &                                     gtu, gsu, gru,  &   ! of t, s, rd at the bottom ocean level
            &                                     gtv, gsv, grv )
      ELSE                                                  ! centered hpg (default case)
                               CALL eos( tb, sb, rhd, rhop )          ! now (swap=before) in situ density for dynhpg module
         IF( ln_zps    )       CALL zps_hde( kstp, tb, sb, rhd,   &   ! Partial steps: now horizontal gradient
            &                                     gtu, gsu, gru,  &   ! of t, s, rd at the bottom ocean level
            &                                     gtv, gsv, grv )
      ENDIF

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! saving ta and sa (temporary fix, please do not remove)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                               zta_tmp(:,:,:) = ta(:,:,:)
                               zsa_tmp(:,:,:) = sa(:,:,:)

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Dynamics
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ta, sa arrays are used as workspace in this section 
      !-----------------------------------------------------------------------
                               ua(:,:,:) = 0.e0               ! set dynamics trends to zero
                               va(:,:,:) = 0.e0

      IF(  ln_asmiau .AND. &
         & ln_dyninc     )     CALL dyn_asm_inc( kstp )       ! apply dynamics assimilation increment

                               CALL dyn_adv( kstp )           ! advection (vector or flux form)
                               CALL dyn_vor( kstp )           ! vorticity term including Coriolis
                               CALL dyn_ldf( kstp )           ! lateral mixing
#if defined key_agrif
      IF(.NOT. Agrif_Root())   CALL Agrif_Sponge_dyn          ! momemtum sponge
#endif
                               CALL dyn_hpg( kstp )           ! horizontal gradient of Hydrostatic pressure
                               CALL dyn_zdf( kstp )           ! vertical diffusion
      IF( lk_dynspg_rl ) THEN
         IF( lk_obc    )       CALL obc_spg( kstp )           ! surface pressure gradient at open boundaries
      ENDIF
                               indic=0
                               CALL dyn_spg( kstp, indic )    ! surface pressure gradient
                               CALL dyn_nxt( kstp )           ! lateral velocity at next time step
      IF( lk_vvl )             CALL dom_vvl                   ! vertical mesh at next time step


      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Computation of diagnostic variables
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------
                       CALL div_cur( kstp )                 ! Horizontal divergence & Relative vorticity
      IF( n_cla == 1 ) CALL div_cla( kstp )                 ! Cross Land Advection (Update Hor. divergence)
                       CALL wzv( kstp )                     ! Vertical velocity

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Control and restarts
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                                 CALL stp_ctl( kstp, indic )
      IF( indic < 0          )   CALL ctl_stop( 'step: indic < 0' )
      IF( kstp == nit000     )   CALL iom_close( numror )             ! close input  ocean restart file
      IF( lrst_oce           )   CALL rst_write  ( kstp )             ! write output ocean restart file
      IF( lk_obc             )   CALL obc_rst_write( kstp )           ! write open boundary restart file

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! diagnostics and outputs
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------

      IF( nstop == 0 ) THEN                                 ! Diagnostics:
         IF( lk_floats  )   CALL flo_stp( kstp )                 ! drifting Floats
         IF( lk_trddyn  )   CALL trd_dwr( kstp )                 ! trends: dynamics 
         IF( lk_trdtra  )   CALL trd_twr( kstp )                 ! trends: active tracers
         IF( lk_trdmld  )   CALL trd_mld( kstp )                 ! trends: Mixed-layer 
         IF( lk_trdvor  )   CALL trd_vor( kstp )                 ! trends: vorticity budget
         IF( lk_diaspr  )   CALL dia_spr( kstp )                 ! Surface pressure diagnostics
         IF( lk_diahth  )   CALL dia_hth( kstp )                 ! Thermocline depth (20 degres isotherm depth)
         IF( lk_diagap  )   CALL dia_gap( kstp )                 ! basin averaged diagnostics
         IF( lk_diahdy  )   CALL dia_hdy( kstp )                 ! dynamical heigh diagnostics
         IF( lk_diafwb  )   CALL dia_fwb( kstp )                 ! Fresh water budget diagnostics
         IF( lk_diaobs  )   CALL dia_obs( kstp )                 ! obs-minus-model (assimilation) diagnostics
         IF( ln_diaptr  )   CALL dia_ptr( kstp )                 ! Poleward TRansports diagnostics

         !                                                 ! Outputs
 

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! restoring ta and sa (temporary fix, please do not remove)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                            ta(:,:,:) = zta_tmp(:,:,:)
                            sa(:,:,:) = zsa_tmp(:,:,:)

         !                                                 ! Outputs
                            CALL dia_wri    ( kstp, indic )      ! ocean model: outputs

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Assimilation mode
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

         IF( lk_asminc ) THEN
            IF( ln_bkgwri ) CALL asm_bkg_wri( kstp )       ! Output background fields
            IF( ln_trjwri ) CALL tam_trj_wri( kstp )       ! Output trajectory fields
            IF( ln_asmdin ) THEN                           ! Direct initialization
               IF( ln_trainc ) CALL tra_asm_inc( kstp )        ! Tracers
               IF( ln_dyninc ) CALL dyn_asm_inc( kstp )        ! Dynamics
               IF( ln_sshinc ) CALL ssh_asm_inc( kstp )        ! SSH
            ENDIF
         ENDIF

      ENDIF

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Coupled mode
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

#if defined key_oasis3
      IF( lk_cpl    )   CALL cpl_stp( kstp )                 ! coupled mode : field exchanges
#endif
      !
   END SUBROUTINE stp_cpd

   !!======================================================================
END MODULE step_cpd
