MODULE step_tam
#ifdef key_tam
   !!======================================================================
   !!                       ***  MODULE step_tam  ***
   !! Time-stepping    : manager of the adjoint ocean time stepping
   !!                    Tangent and Adjoint module
   !!======================================================================
   !! History of 
   !! the direct:      !  91-03  ()  Original code
   !!                  !  91-11  (G. Madec)
   !!                  !  92-06  (M. Imbard)  add a first output record
   !!                  !  96-04  (G. Madec)  introduction of dynspg
   !!                  !  96-04  (M.A. Foujols)  introduction of passive tracer
   !!             8.0  !  97-06  (G. Madec)  new architecture of call
   !!             8.2  !  97-06  (G. Madec, M. Imbard, G. Roullet)  free surface
   !!             8.2  !  99-02  (G. Madec, N. Grima)  hpg implicit
   !!             8.2  !  00-07  (J-M Molines, M. Imbard)  Open Bondary Conditions
   !!             9.0  !  02-06  (G. Madec)  free form, suppress macro-tasking
   !!             " "  !  04-08  (C. Talandier) New trends organization
   !!             " "  !  05-01  (C. Ethe) Add the KPP closure scheme
   !!             " "  !  05-11  (V. Garnier) Surface pressure gradient organization
   !!             " "  !  05-11  (G. Madec)  Reorganisation of tra and dyn calls
   !!             " "  !  06-01  (L. Debreu, C. Mazauric)  Agrif implementation
   !!             " "  !  06-07  (S. Masson)  restart using iom
   !!    "        " "  !  07-04  (K. Mogensen, A. Weaver, M. Martin) Assimilation interface
   !! History of the TAM
   !!                  !  08-06  (A. Vidard and A. Weaver) Tangent and Adjoint version of 9.0
   !!                  !  08-11  (A. Vidard) Nemo v3 update
   !!                  !  06-09  (F. Vigilant)  Modified to split NEMOVAR / NEMOTAM
   !!----------------------------------------------------------------------
   !!   stp_tam        : OPA system time-stepping (tangent linear)
   !!   stp_adj        : OPA system time-stepping (adjoint)
   !!----------------------------------------------------------------------

   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpi,                 &
      & jpj,                 & 
      & jpk,                 &
      & jpiglo
   USE oce           , ONLY: & ! ocean dynamics and tracers variables
      & tb, sb, tn, sn, ta,  &
      & un, vn, sshn, sshb,  &
      & sa, ub, vb, rn2,     &
      & ln_dynhpg_imp
   USE zdfkpp        , ONLY: &
      & lk_zdfkpp      ! KPP vertical mixing flag
   USE dynspg_oce    , ONLY: &
      &  lk_dynspg_rl  ! Rigid-lid flag
   USE dom_oce       , ONLY: & ! ocean space and time domain variables 
      & n_cla, e1u, e2u,     &
      & e1v, e2v, e1t, e2t,  &
# if defined key_vvl
      & e3t_1,               &
# else
#  if defined key_zco
      & e3t_0, e3w_0,        &
#  else
      & e3t, e3u, e3v, e3w,  &
#  endif
# endif
      & nldi, nldj, nlei,    &
      & nlej, lk_vvl, ln_zps,&
      & narea, mig, mjg,     &
      & umask, vmask,  tmask
   USE in_out_manager, ONLY: & ! I/O manager
      & numout, lwp,         &
      & nit000, nitend,      &
      & nstop, ctl_stop
   USE daymod        , ONLY: &
      & adatrj
    USE trabbc        , ONLY: &! bottom boundary condition
      & lk_trabbc
   USE traqsr        , ONLY: &
      &  ln_traqsr
          ! solar radiation penetration flag
   USE asminc
   USE asmbkg
   USE oce_tam       , ONLY: & ! Tangent linear and adjoint variables
      & oce_tam_init,        &
      & un_tl, vn_tl, tn_tl, &
      & sn_tl, ua_tl, va_tl, &
      & ta_tl, sa_tl, ub_tl, &
      & vb_tl, tb_tl, sb_tl, & 
      & un_ad, vn_ad, tn_ad, &
      & sn_ad, ua_ad, va_ad, &
      & ta_ad, sa_ad, ub_ad, &
      & vb_ad, tb_ad, sb_ad, & 
      & sshn_tl, sshn_ad,    &
#if defined key_obc
      & sshb_ad,             &
#endif
      & rn2_tl, rn2_ad,      &
      & rhd_tl, rhop_tl,     &
      & rhd_ad, rhop_ad,     &
      & gtu_tl, gsu_tl,      &
      & gru_tl, gtv_tl,      &
      & gsv_tl, grv_tl,      &
      & gtu_ad, gsu_ad,      &
      & gru_ad, gtv_ad,      &
      & gsv_ad, grv_ad,      &
      & ssha_tl, sshb_tl,    &
      & rotb_tl, hdivb_tl,   &
      & rotn_tl, hdivn_tl,   &
      & wn_tl
   USE lbclnk_tam
   USE daymod_tam      ! calendar                         (adjoint of day     routine)
   USE sbc_oce_tam
   USE sbcmod_tam
   USE traqsr_tam      ! solar radiation penetration      (adjoint of tra_qsr routine)
   USE trasbc_tam      ! surface boundary condition       (adjoint of tra_sbc routine)
   USE trabbc_tam      ! bottom boundary condition        (adjoint of tra_bbc routine)
   USE tradmp_tam      ! internal damping                 (adjoint of tra_dmp routine)
   USE traadv_tam      ! advection scheme control     (adjoint of tra_adv_ctl routine)
   USE traldf_tam      ! lateral mixing                   (adjoint of tra_ldf routine)
   USE cla_tam         ! cross land advection             (adjoint of tra_cla routine)
   !   zdfkpp          ! KPP non-local tracer fluxes      (adjoint of tra_kpp routine)
   USE trazdf_tam      ! vertical mixing                  (adjoint of tra_zdf routine)
   USE tranxt_tam      ! time-stepping                    (adjoint of tra_nxt routine)
   USE eosbn2_tam      ! equation of state                (adjoint of eos_bn2 routine)

   USE dynadv_tam      ! advection                        (adjoint of dyn_adv routine)
   USE dynvor_tam      ! vorticity term                   (adjoint of dyn_vor routine)
   USE dynhpg_tam      ! hydrostatic pressure grad.       (adjoint of dyn_hpg routine)
   USE dynldf_tam      ! lateral momentum diffusion       (adjoint of dyn_ldf routine)
   USE dynzdf_tam      ! vertical diffusion               (adjoint of dyn_zdf routine)
   USE dynspg_tam      ! surface pressure gradient        (adjoint of dyn_spg routine)
   USE dynnxt_tam      ! time-stepping                    (adjoint of dyn_nxt routine)

!   USE bdy_par_tam
!   USE bdydta_tam

   USE divcur_tam      ! hor. divergence and curl      (adjoint of div & cur routines)
   USE cla_div_tam     ! cross land: hor. divergence      (adjoint of div_cla routine)
   USE wzvmod_tam      ! vertical velocity                (adjoint of wzv     routine)

!!   USE zdfkpp_tam     ! KPP vertical mixing
   USE zdf_oce, ONLY : lk_zdfcst, avt, avt0, avmu, avmv, avm0, ln_zdfevd         ! KPP vertical mixing
   USE zdfddm, ONLY  : &            ! double diffusion mixing          (zdf_ddm routine)
      & lk_zdfddm,     &
      & zdf_ddm  
   USE zdfevd, only:zdf_evd         ! double diffusion mixing          (zdf_ddm routine)
   USE zdfbfr, only:zdf_bfr         ! double diffusion mixing          (zdf_ddm routine)
   USE zdfmxl, only:zdf_mxl         ! double diffusion mixing          (zdf_ddm routine)
   USE eosbn2, ONLY: bn2
   USE zpshde_tam      ! partial step: hor. derivative     (adjoint of zps_hde routine)

!!   USE diaobs_tam      ! obs-minus-model (adjoint of assimilation)   (adjoint of dia_obs routine)

   USE trj_tam

   USE stpctl_tam      ! time stepping control            (adjoint of stp_ctl routine)

   USE gridrandom, ONLY: & 
                       ! Random Gaussian noise on grids
      & grid_random,     &
      & grid_rd_sd
   USE dotprodfld,     ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE tstool_tam, ONLY: &
      & prntst_adj,      & 
      & prntst_tlm,      &
      & stdemp,          & 
      & stdu,stdv,       &
      & stdt,            &   
      & stds, stdssh,    &
      & stdr

   USE paresp, ONLY:     & 
                       ! Weights for an energy-type scalar product
      & wesp_t,          & 
      & wesp_s,          & 
      & wesp_u,          & 
      & wesp_ssh           

   USE istate_tam      !: Initial state setting          (istate_init routine)
   USE sol_oce_tam
   USE trc_oce_tam

   USE step
   USE iom ! for dumping an abort state 
#if defined key_obc
   USE obc_par, ONLY : lk_obc
   USE obcdta          ! open boundary condition data     (obc_dta routine)
#endif

#if defined key_agrif
#error 'agrif not yet implemented in nemotam'
#endif

   IMPLICIT NONE
   PRIVATE

   PUBLIC stp_tan,      &
      &   stp_adj,      & ! called by simvar.F90
#if defined key_tst_tlm
      &   stp_tlm_tst,  &
#endif
      &   stp_adj_tst

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "zdfddm_substitute.h90"

CONTAINS
#if defined key_agrif
   #error 'agrif not yet implemented in nemotam'
  ! SUBROUTINE stp_tan( )
#else
   SUBROUTINE stp_tan( kstp )
#endif
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE stp_tan  ***
      !!                      
      !! ** Purpose of the direct routine: 
      !!              - Time stepping of OPA (momentum and active tracer eqs.)
      !!              - Time stepping of LIM (dynamic and thermodynamic eqs.)
      !!              - Tme stepping  of TRC (passive tracer eqs.)
      !! 
      !! ** Method of the direct routine:
      !!              -1- Update forcings and data  
      !!              -2- Update ocean physics 
      !!              -3- Compute the t and s trends 
      !!              -4- Update t and s 
      !!              -5- Compute the momentum trends
      !!              -6- Update the horizontal velocity
      !!              -7- Compute the diagnostics variables (rd,N2, div,cur,w)
      !!              -8- Outputs and diagnostics
      !!----------------------------------------------------------------------
      !! * Arguments
#if defined key_agrif   
      INTEGER               :: kstp   ! ocean time-step index
#else
      INTEGER, INTENT( in ) :: kstp   ! ocean time-step index
#endif      

      !! * local declarations
      INTEGER ::   indic    ! error indicator if < 0
      REAL(wp), DIMENSION(jpi,jpj,jpk) :: zta_tmp, zsa_tmp
      !! ---------------------------------------------------------------------

#if defined key_agrif
      kstp = nit000 + Agrif_Nb_Step()
      !      IF ( Agrif_Root() .and. lwp) Write(*,*) '---'
      !      IF (lwp) Write(*,*) 'Grid Number',Agrif_Fixed(),' time step ',kstp
#endif   
      indic = 1                    ! reset to no error condition

      CALL day_tam( kstp, 0 )             ! Calendar

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Update data, open boundaries, surface boundary condition (including sea-ice)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

      ! update 3D temperature data ... not needed in tangent

      ! update 3D salinity data ... not needed in tangent (to be investigated, see sbc_ssr)

      CALL sbc_tan ( kstp ) ! Sea boundary condition (including sea-ice) 

!      IF( lk_obc     )   CALL obc_dta( kstp )  ! ???       ! update dynamic and tracer data at open boundaries
      ! update dynamic and tracer data at open boundaries ... not needed for the time being, to investigate whenever we do obc.

      ! compute phase velocities at open boundaries ... not needed for the time being, to investigate whenever we do obc.
!      IF( lk_bdy     )   CALL bdy_dta_tan( kstp )         ! update dynamic and tracer data at unstructured open boundary

      ! Output the initial state and forcings ... not needed in tangent

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Ocean physics update
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      !-----------------------------------------------------------------------
      !  VERTICAL PHYSICS
      !-----------------------------------------------------------------------
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------

      CALL bn2_tan( tb, sb, tb_tl, sb_tl, rn2_tl )              ! before Brunt-Vaisala frequency

      !                                                     ! Vertical eddy viscosity and diffusivity coefficients
      ! Richardson number dependent Kz  ... not available
      ! TKE closure scheme for Kz ... not available
      ! KPP closure scheme for Kz ... not available

      ! Constant Kz (reset avt, avm[uv] to the background value)... not available

      ! lk_zdfcst_tan:  Constant Kz read from the reference trajectory


      ! ! increase diffusivity at rivers mouths... not needed in tangent 

      ! enhanced vertical eddy diffusivity ... not needed in tangent with lk_zdfcst_tan

      ! double diffusive mixing ... not needed in tangent with lk_zdfcst_tan

      ! bottom friction... not needed in tangent with lk_zdfcst_tan

      ! mixed layer depth... not needed in tangent with lk_zdfcst_tan


      !-----------------------------------------------------------------------
      !  LATERAL PHYSICS
      !-----------------------------------------------------------------------
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------

      ! before slope of the lateral mixing... not needed in tangent with lk_zdfcst_tan

#if defined key_traldf_c2d
      ! eddy induced velocity coefficient... not needed in tangent with lk_zdfcst_tan
#endif


#if defined key_top
      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Passive Tracer Model
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------

      ! time-stepping... not needed in tangent for the time being

#endif
      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Active tracers
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va arrays are used as workspace in this section
      !-----------------------------------------------------------------------

      ta_tl(:,:,:) = 0.e0            ! set tracer trends to zero
      sa_tl(:,:,:) = 0.e0

      ! apply tracer assimilation increment ... not needed in tangent
      ! apply bias directly to tn, sn, tb, sb ... not needed in tangent
      CALL tra_sbc_tan( kstp )       ! surface boundary condition

      IF( ln_traqsr      )   CALL tra_qsr_tan( kstp )       ! penetrative solar radiation qsr

      IF( lk_trabbc      )   CALL tra_bbc_tan( kstp )       ! bottom heat flux

      ! diffusive bottom boundary layer scheme ... currently not available
      ! advective (and/or diffusive) bottom boundary layer scheme ... currently not available

      IF( lk_tradmp      )   CALL tra_dmp_tan( kstp )       ! internal damping trends

      CALL tra_adv_tan( kstp )       ! horizontal & vertical advection

      IF( n_cla == 1     )   CALL tra_cla_tan( kstp )       ! Cross Land Advection (Update Hor. advection)

!      IF( lk_zdfkpp )        CALL tra_kpp_tan( kstp )       ! KPP non-local tracer fluxes

      CALL tra_ldf_tan( kstp )       ! lateral mixing

#if defined key_agrif
      ! tracers sponge ... not available
#endif
      CALL tra_zdf_tan( kstp )       ! vertical mixing

      ! update the new (t,s) fields by non
      ! penetrative convective adjustment ... not available 

      CALL tra_nxt_tan( kstp )           ! tracer fields at next time step

      IF( ln_dynhpg_imp  ) THEN                             ! semi-implicit hpg
         CALL eos_tan( ta, sa, ta_tl, sa_tl, rhd_tl, rhop_tl )      ! Time-filtered in situ density used in dynhpg module
         IF( ln_zps    )       CALL zps_hde_tan( kstp, ta, sa, ta_tl, sa_tl, rhd_tl,&   ! Partial steps: time filtered hor. gradient
            &                                     gtu_tl, gsu_tl, gru_tl,   &   ! of t, s, rd at the bottom ocean level
            &                                     gtv_tl, gsv_tl, grv_tl )
      ELSE                                                  ! centered hpg (default case)
         CALL eos_tan( tb, sb, tb_tl, sb_tl, rhd_tl, rhop_tl )       ! now (swap=before) in situ density for dynhpg module
         IF( ln_zps    )       CALL zps_hde_tan( kstp, tb, sb, tb_tl, sb_tl, rhd_tl,&   ! Partial steps: now horizontal gradient
            &                                     gtu_tl, gsu_tl, gru_tl,   &   ! of t, s, rd at the bottom ocean level
            &                                     gtv_tl, gsv_tl, grv_tl )
      ENDIF

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Dynamics
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ta, sa arrays are used as workspace in this section 
      !-----------------------------------------------------------------------

      ua_tl(:,:,:) = 0.e0               ! set dynamics trends to zero
      va_tl(:,:,:) = 0.e0

      ! apply dynamics assimilation increment ... not needed in tangent

      CALL dyn_adv_tan( kstp )           ! advection (vector or flux form)
      CALL dyn_vor_tan( kstp )           ! vorticity term including Coriolis
      CALL dyn_ldf_tan( kstp )           ! lateral mixing
#if defined key_agrif
      ! momemtum sponge ... not available
#endif

      CALL dyn_hpg_tan( kstp )           ! horizontal gradient of Hydrostatic pressure
      CALL dyn_zdf_tan( kstp )           ! vertical diffusion

      IF( lk_dynspg_rl ) THEN
         ! surface pressure gradient at open boundaries ... not available
      ENDIF
      indic=0
      !i
      CALL dyn_spg_tan( kstp, indic )    ! surface pressure gradient
      CALL dyn_nxt_tan( kstp )           ! lateral velocity at next time step

      ! vertical mesh at next time step ... not available

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Computation of diagnostic variables
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------

      CALL div_cur_tan( kstp )                 ! Horizontal divergence & Relative vorticity

      IF( n_cla == 1 ) CALL div_cla_tan( kstp )                 ! Cross Land Advection (Update Hor. divergence)

      CALL wzv_tan( kstp )                     ! Vertical velocity

      CALL trj_rea( kstp, 1) ! ... Read basic state trajectory at end of current step

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Control, and restarts
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------
      !                                                     ! Time loop: control and print
      CALL stp_ctl_tan( kstp, indic, 0 )
      IF( indic < 0          )   CALL ctl_stop( 'step_tan: indic < 0' )

      ! close input  ocean restart file ... not needed in tangent
      ! write output ocean restart file... not needed in tangent
      ! write open boundary restart file... not needed in tangent

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! diagnostics and outputs
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------

      IF ( nstop == 0 ) THEN                                ! Diagnostics
         ! drifting Floats... not needed in tangent
         ! trends: dynamics ... not needed in tangent
         ! trends: active tracers... not needed in tangent
         ! trends: Mixed-layer ... not needed in tangent
         ! trends: vorticity budget... not needed in tangent
         ! Surface pressure diagnostics... not needed in tangent
         ! Thermocline depth (20 degres isotherm depth)... not needed in tangent
         ! basin averaged diagnostics... not needed in tangent
         ! dynamical heigh diagnostics... not needed in tangent
         ! Fresh water budget diagnostics... not needed in tangent
!!!         IF( lk_diaobs  )   CALL dia_obs_tan( kstp )                 ! obs-minus-model (assimilation) diagnostics NOTE: may be better off outside this module
         ! Poleward TRansports diagnostics... not needed in tangent


         !                                                 ! Outputs
         ! ocean model: outputs... not needed in tangent

      ENDIF

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Coupled mode
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

      ! coupled mode : field exchanges ... not available for the next 20 years
      !
   END SUBROUTINE stp_tan

#if defined key_agrif
#error 'agrif not yet implemented in nemotam'
   ! SUBROUTINE stp_adj( )
#else
   SUBROUTINE stp_adj( kstp )
#endif
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE stp_adj  ***
      !!                      
      !! ** Purpose of the direct routine: 
      !!              - Time stepping of OPA (momentum and active tracer eqs.)
      !!              - Time stepping of LIM (dynamic and thermodynamic eqs.)
      !!              - Tme stepping  of TRC (passive tracer eqs.)
      !! 
      !! ** Method of the direct routine:
      !!              -1- Update forcings and data  
      !!              -2- Update ocean physics 
      !!              -3- Compute the t and s trends 
      !!              -4- Update t and s 
      !!              -5- Compute the momentum trends
      !!              -6- Update the horizontal velocity
      !!              -7- Compute the diagnostics variables (rd,N2, div,cur,w)
      !!              -8- Outputs and diagnostics
      !!----------------------------------------------------------------------
      !! * Arguments
#if defined key_agrif   
      INTEGER               :: kstp   ! ocean time-step index
#else
      INTEGER, INTENT( in ) :: kstp   ! ocean time-step index
#endif      

      REAL(wp), DIMENSION(jpi,jpj,jpk) :: zta_tmp, zsa_tmp

      !! * local declarations
      INTEGER ::   indic    ! error indicator if < 0
      !! ---------------------------------------------------------------------

#if defined key_agrif
      kstp = nit000 + Agrif_Nb_Step()
      !      IF ( Agrif_Root() .and. lwp) Write(*,*) '---'
      !      IF (lwp) Write(*,*) 'Grid Number',Agrif_Fixed(),' time step ',kstp
#endif   
      indic = 1                    ! reset to no error condition

      CALL day_tam( kstp, 0 )             ! Calendar

      ! ... Read basic state trajectory at end of previous step
      CALL trj_rea( ( kstp - 1 ), -1 )

      zta_tmp (:,:,:) = ta (:,:,:) 
      zsa_tmp (:,:,:) = sa (:,:,:) 
      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Computation of diagnostic variables
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------

      ! ocean surface freezing temperature ... not needed in adjoint

      CALL wzv_adj( kstp )                     ! Vertical velocity
      IF( n_cla == 1 ) CALL div_cla_adj( kstp )                 ! Cross Land Advection (Update Hor. divergence)


      CALL div_cur_adj( kstp )                 ! Horizontal divergence & Relative vorticity
      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Dynamics
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ta, sa arrays are used as workspace in this section 
      !-----------------------------------------------------------------------
      ! apply dynamics assimilation increment ... not needed in adjoint
      indic=0
      !i bug lbc sur emp

      ! vertical mesh at next time step ... not available
      CALL dyn_nxt_adj( kstp )           ! lateral velocity at next time step
      CALL dyn_spg_adj( kstp, indic )    ! surface pressure gradient
      !i
      IF( lk_dynspg_rl ) THEN
         ! surface pressure gradient at open boundaries ... not available
      ENDIF
      CALL dyn_zdf_adj( kstp )           ! vertical diffusion
      CALL dyn_hpg_adj( kstp )           ! horizontal gradient of Hydrostatic pressure
#if defined key_agrif
      ! momemtum sponge ... not available
#endif
      CALL dyn_ldf_adj( kstp )           ! lateral mixing
      CALL dyn_vor_adj( kstp )           ! vorticity term including Coriolis
      CALL dyn_adv_adj( kstp )           ! advection (vector or flux form)

      ta (:,:,:) = zta_tmp (:,:,:)
      sa (:,:,:) = zsa_tmp (:,:,:)

      ua_ad(:,:,:) = 0.e0                ! set tracer trends to zero
      va_ad(:,:,:) = 0.e0

      IF( ln_dynhpg_imp  ) THEN                             ! semi-implicit hpg
         IF( ln_zps    )       CALL zps_hde_adj( kstp, ta, sa, ta_ad, sa_ad, rhd_ad,&   ! Partial steps: time filtered hor. gradient
            &                                     gtu_ad, gsu_ad, gru_ad,   &   ! of t, s, rd at the bottom ocean level
            &                                     gtv_ad, gsv_ad, grv_ad )
         CALL eos_adj( ta, sa, ta_ad, sa_ad, rhd_ad, rhop_ad )      ! Time-filtered in situ density used in dynhpg module
      ELSE                                                  ! centered hpg (default case)
         IF( ln_zps    )       CALL zps_hde_adj( kstp, tb, sb, &
            &                                     tb_ad, sb_ad,rhd_ad,&   ! Partial steps: now horizontal gradient
            &                                     gtu_ad, gsu_ad, gru_ad,   &   ! of t, s, rd at the bottom ocean level
            &                                     gtv_ad, gsv_ad, grv_ad )
         CALL eos_adj( tb, sb, tb_ad, sb_ad, rhd_ad, rhop_ad )       ! now (swap=before) in situ density for dynhpg module
      ENDIF

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Active tracers
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va arrays are used as workspace in this section
      !-----------------------------------------------------------------------

      CALL tra_nxt_adj( kstp )           ! tracer fields at next time step
      ! update the new (t,s) fields by non
      ! penetrative convective adjustment ... not available 
      CALL tra_zdf_adj( kstp )       ! vertical mixing

#if defined key_agrif
      ! tracers sponge ... not available
#endif

      CALL tra_ldf_adj( kstp )       ! lateral mixing

!      IF( lk_zdfkpp )        CALL tra_kpp_adj( kstp )       ! KPP non-local tracer fluxes

      IF( n_cla == 1     )   CALL tra_cla_adj( kstp )       ! Cross Land Advection (Update Hor. advection)

      CALL tra_adv_adj( kstp )       ! horizontal & vertical advection

      IF( lk_tradmp      )   CALL tra_dmp_adj( kstp )       ! internal damping trends
      ! advective (and/or diffusive) bottom boundary layer scheme ... currently not available
      ! diffusive bottom boundary layer scheme ... currently not available

      IF( lk_trabbc      )   CALL tra_bbc_adj( kstp )       ! bottom heat flux

      IF( ln_traqsr      )   CALL tra_qsr_adj( kstp )       ! penetrative solar radiation qsr

      CALL tra_sbc_adj( kstp )       ! surface boundary condition

      ta_ad(:,:,:) = 0.e0            ! set tracer trends to zero
      sa_ad(:,:,:) = 0.e0
      ! apply tracer assimilation increment ... not needed in adjoint
#if defined key_top
      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Passive Tracer Model
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------

      ! time-stepping... not needed in adjoint for the time being

#endif
      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Ocean physics update
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      !-----------------------------------------------------------------------
      !  LATERAL PHYSICS
      !-----------------------------------------------------------------------
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------
#if defined key_traldf_c2d
      ! eddy induced velocity coefficient... not needed in tangent with lk_zdfcst_adj
#endif
      ! before slope of the lateral mixing... not needed in adjoint with lk_zdfcst_adj
      !-----------------------------------------------------------------------
      !  VERTICAL PHYSICS
      !-----------------------------------------------------------------------
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------
      ! mixed layer depth... not needed in adjoint with lk_zdfcst_adj
      ! bottom friction... not needed in adjoint with lk_zdfcst_adj
      ! double diffusive mixing ... not needed in adjoint with lk_zdfcst_adj
      ! enhanced vertical eddy diffusivity ... not needed in adjoint with lk_zdfcst_adj
      ! ! increase diffusivity at rivers mouths... not needed in tangent 
      ! lk_zdfcst_adj:  Constant Kz read from the reference trajectory
      ! Constant Kz (reset avt, avm[uv] to the background value)... not available
      ! KPP closure scheme for Kz ... not available
      ! TKE closure scheme for Kz ... not available
      ! Richardson number dependent Kz  ... not available
      !                                                     ! Vertical eddy viscosity and diffusivity coefficients
      CALL bn2_adj( tb, sb, tb_ad, sb_ad, rn2_ad )              ! before Brunt-Vaisala frequency

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Update data, open boundaries and Forcings
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! Output the initial state and forcings ... not needed in adjoint
!      IF( lk_bdy     )   CALL bdy_dta_adj( kstp )         ! update dynamic and tracer data at unstructured open boundary
      ! compute phase velocities at open boundaries ... not needed for the time being, to investigate whenever we do obc.
      ! update dynamic and tracer data at open boundaries ... not needed for the time being, to investigate whenever we do obc.

      CALL sbc_adj ( kstp ) ! Sea boundary condition (including sea-ice) 
      ! update 3D salinity data ... not needed in tangent
      ! update 3D temperature data ... not needed in adjoint

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Control, and restarts
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------
      !                                                     ! Time loop: control and print
      CALL stp_ctl_adj( kstp, indic, 0 )
      IF( indic < 0          )   CALL ctl_stop( 'step_adj: indic < 0' )

      ! close input  ocean restart file ... not needed in adjoint
      ! write output ocean restart file... not needed in adjoint
      ! write open boundary restart file... not needed in adjoint

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! diagnostics and outputs
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      ! N.B. ua, va, ta, sa arrays are used as workspace in this section
      !-----------------------------------------------------------------------

      IF ( nstop == 0 ) THEN                                ! Diagnostics
         ! drifting Floats... not needed in adjoint
         ! trends: dynamics ... not needed in adjoint
         ! trends: active tracers... not needed in adjoint
         ! trends: Mixed-layer ... not needed in adjoint
         ! trends: vorticity budget... not needed in adjoint
         ! Surface pressure diagnostics... not needed in adjoint
         ! Thermocline depth (20 degres isotherm depth)... not needed in adjoint
         ! basin averaged diagnostics... not needed in adjoint
         ! dynamical heigh diagnostics... not needed in adjoint
         ! Fresh water budget diagnostics... not needed in adjoint
!!!         IF( lk_diaobs  )   CALL dia_obs_adj( kstp )                 ! obs-minus-model (assimilation) diagnostics NOTE: may be better off outside this module
         ! Poleward TRansports diagnostics... not needed in adjoint


         !                                                 ! Outputs
         ! ocean model: outputs... not needed in adjoint

      ENDIF

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Coupled mode
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

      ! coupled mode : field exchanges ... not available for the next 20 years
      !
   END SUBROUTINE stp_adj

   SUBROUTINE stp_adj_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE example_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !!                   
      !! History :
      !!        ! 09-03 (F. Vigilant)
      !!-----------------------------------------------------------------------
      !! * Modules used
      USE sbcssr_tam, ONLY: &
         & qrp_tl, qrp_ad,  &
         & erp_tl, erp_ad
      USE sbcfwb_tam, ONLY: &
         & a_fwb_tl,        & ! for before year.
         & a_fwb_ad  ! for before year.

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt               ! Output unit
 
      !! * Local declarations
      INTEGER ::             &
         & ji,               &  ! dummy loop indices
         & jj,               &        
         & jk,               &
         & istp
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d             ! 2D seed for the random number generator
      REAL(KIND=wp) ::       &
         & zsp1    ,         &  ! scalar product involving the tangent routine
         & zsp1_U  ,         &
         & zsp1_V  ,         & 
         & zsp1_T  ,         & 
         & zsp1_S  ,         & 
         & zsp1_SSH,         &   
         & zsp2    ,         &  ! scalar product involving the adjoint routine
         & zsp2_U  ,         &
         & zsp2_V  ,         & 
         & zsp2_T  ,         & 
         & zsp2_S  ,         & 
         & zsp2_SSH     
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & zun_tlin ,        & ! Tangent input
         & zvn_tlin ,        & ! Tangent input
         & ztn_tlin ,        & ! Tangent input
         & zsn_tlin ,        & ! Tangent input
         & zun_tlout,        & ! Tangent output
         & zvn_tlout,        & ! Tangent output
         & ztn_tlout,        & ! Tangent output
         & zsn_tlout,        & ! Tangent outpu
         & zun_adin ,        & ! Adjoint input
         & zvn_adin ,        & ! Adjoint input
         & ztn_adin ,        & ! Adjoint input
         & zsn_adin ,        & ! Adjoint input
#if defined key_obc
         & zub_adin ,        & ! Adjoint input
         & zvb_adin ,        & ! Adjoint input
         & ztb_adin ,        & ! Adjoint input
         & zsb_adin ,        & ! Adjoint input
         & zub_tlout ,        & ! Adjoint input
         & zvb_tlout ,        & ! Adjoint input
         & ztb_tlout ,        & ! Adjoint input
         & zsb_tlout ,        & ! Adjoint input
#endif
         & zun_adout,        & ! Adjoint output
         & zvn_adout,        & ! Adjoint output
         & ztn_adout,        & ! Adjoint output
         & zsn_adout,        & ! Adjoint output
         & z3r                 ! 3D random field  
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zsshn_tlin  ,     & ! Tangent input
         & zsshn_tlout ,     & ! Tangent input
         & zsshn_adin  ,     & ! Adjoint output
         & zsshn_adout ,     & ! Adjoint output
#if defined key_obc
         & zsshb_tlout ,     & ! Tangent input
         & zsshb_adin  ,     & ! Tangent input
#endif
         & z2r                 ! 2D random field 

      CHARACTER(LEN=14) ::   &
         & cl_name

      INTEGER ::             &
         & jpert
      INTEGER, PARAMETER :: jpertmax = 6

      ! Allocate memory

      ALLOCATE( &
         & zun_tlin(   jpi,jpj,jpk), zvn_tlin(   jpi,jpj,jpk), ztn_tlin(  jpi,jpj,jpk), &
         & zsn_tlin(   jpi,jpj,jpk), zsshn_tlin( jpi,jpj    ), zun_tlout( jpi,jpj,jpk), &
         & zvn_tlout(  jpi,jpj,jpk), ztn_tlout(  jpi,jpj,jpk), zsn_tlout( jpi,jpj,jpk), &
         & zsshn_tlout(jpi,jpj    ), zun_adin(   jpi,jpj,jpk), zvn_adin(  jpi,jpj,jpk), &
         & ztn_adin(   jpi,jpj,jpk), zsn_adin(   jpi,jpj,jpk), zsshn_adin(jpi,jpj    ), &
         & zun_adout(  jpi,jpj,jpk), zvn_adout(  jpi,jpj,jpk), ztn_adout( jpi,jpj,jpk), &
         & zsn_adout(  jpi,jpj,jpk), zsshn_adout(jpi,jpj    ), z2r(       jpi,jpj    ), &
         & z3r(        jpi,jpj,jpk)                                                     &
         & )
    
#if defined key_obc
      ALLOCATE( zub_adin   (jpi,jpj,jpk), zvb_adin  (jpi,jpj,jpk) ,        &
           &    ztb_adin   (jpi,jpj,jpk), zsb_adin  (jpi,jpj,jpk) ,        &
           &    zub_tlout  (jpi,jpj,jpk), zvb_tlout (jpi,jpj,jpk) ,        &
           &    ztb_tlout  (jpi,jpj,jpk), zsb_tlout (jpi,jpj,jpk) ,        &
           &    zsshb_tlout(jpi,jpj)    , zsshb_adin(jpi,jpj)       )
#endif
      !==================================================================
      ! 1) dx = ( un_tl, vn_tl, tn_tl, sn_tl, sshn_tl ) and 
      !    dy = ( ..... )
      !==================================================================


      DO jpert = 1, jpertmax

         !--------------------------------------------------------------------
         ! Reset the tangent and adjoint variables
         !--------------------------------------------------------------------
         zun_tlin   (:,:,:) = 0.0_wp
         zvn_tlin   (:,:,:) = 0.0_wp
         ztn_tlin   (:,:,:) = 0.0_wp
         zsn_tlin   (:,:,:) = 0.0_wp
         zsshn_tlin (  :,:) = 0.0_wp
         zun_tlout  (:,:,:) = 0.0_wp
         zvn_tlout  (:,:,:) = 0.0_wp
         ztn_tlout  (:,:,:) = 0.0_wp
         zsn_tlout  (:,:,:) = 0.0_wp
         zsshn_tlout(  :,:) = 0.0_wp
         zun_adin   (:,:,:) = 0.0_wp
         zvn_adin   (:,:,:) = 0.0_wp
         ztn_adin   (:,:,:) = 0.0_wp
         zsn_adin   (:,:,:) = 0.0_wp
         zsshn_adin (  :,:) = 0.0_wp
         zun_adout  (:,:,:) = 0.0_wp
         zvn_adout  (:,:,:) = 0.0_wp
         ztn_adout  (:,:,:) = 0.0_wp
         zsn_adout  (:,:,:) = 0.0_wp
         zsshn_adout(  :,:) = 0.0_wp
         z3r        (:,:,:) = 0.0_wp
         z2r        (  :,:) = 0.0_wp

         !--------------------------------------------------------------------
         ! Initialize the tangent input with random noise: dx
         !--------------------------------------------------------------------
       
         IF ( (jpert == 1) .OR. (jpert == jpertmax) ) THEN
            DO jj = 1, jpj
               DO ji = 1, jpi
                  iseed_2d(ji,jj) = - ( 596035 + &
                     &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
               END DO
            END DO
            CALL grid_random( iseed_2d, z3r, 'U', 0.0_wp, stdu )
            DO jk = 1, jpk
               DO jj = nldj, nlej
                  DO ji = nldi, nlei
                     zun_tlin(ji,jj,jk) = z3r(ji,jj,jk)
                  END DO
               END DO
            END DO
         ENDIF
         IF ( (jpert == 2) .OR. (jpert == jpertmax) ) THEN
            DO jj = 1, jpj
               DO ji = 1, jpi
                  iseed_2d(ji,jj) = - ( 371836 + &
                     &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
               END DO
            END DO
            CALL grid_random( iseed_2d, z3r, 'V', 0.0_wp, stdv ) 
            DO jk = 1, jpk
               DO jj = nldj, nlej
                  DO ji = nldi, nlei
                     zvn_tlin(ji,jj,jk) = z3r(ji,jj,jk)
                  END DO
               END DO
            END DO
         ENDIF
         IF ( (jpert == 3) .OR. (jpert == jpertmax) ) THEN
            DO jj = 1, jpj
               DO ji = 1, jpi
                  iseed_2d(ji,jj) = - ( 284035 + &
                     &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
               END DO
            END DO
            CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stdt )
            DO jk = 1, jpk
               DO jj = nldj, nlej
                  DO ji = nldi, nlei
                     ztn_tlin(ji,jj,jk) = z3r(ji,jj,jk)
                  END DO
               END DO
            END DO 
         ENDIF
         IF ( (jpert == 4) .OR. (jpert == jpertmax) ) THEN
            DO jj = 1, jpj
               DO ji = 1, jpi
                  iseed_2d(ji,jj) = - (  471426 + &
                     &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
               END DO
            END DO
            CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stds ) 
            DO jk = 1, jpk
               DO jj = nldj, nlej
                  DO ji = nldi, nlei
                     zsn_tlin(ji,jj,jk) = z3r(ji,jj,jk)
                  END DO
               END DO
            END DO
         ENDIF
         IF ( (jpert == 5) .OR. (jpert == jpertmax) ) THEN
            DO jj = 1, jpj
               DO ji = 1, jpi
                  iseed_2d(ji,jj) = - ( 712035 + &
                     &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
               END DO
            END DO
            CALL grid_random( iseed_2d, z2r, 'T', 0.0_wp, stdssh )
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zsshn_tlin(ji,jj) = z2r(ji,jj)
               END DO
            END DO
         ENDIF
         IF (jpert>0) THEN
         CALL     oce_tam_init( 0 )    ! kindic = 0  allocate/initialize tl and adj variables
         CALL sbc_oce_tam_init( 0 )    ! kindic = 0  allocate/initialize tl and adj variables 
         CALL sol_oce_tam_init( 1 )   
         CALL sol_oce_tam_init( 2 )   
         CALL trc_oce_tam_init( 0 )

         qrp_tl = 0.0_wp
         qrp_ad = 0.0_wp
         erp_tl = 0.0_wp
         erp_ad = 0.0_wp

         emp_tl = 0.e0
         emp_ad = 0.e0

         a_fwb_tl = 0.0_wp
         a_fwb_ad = 0.0_wp
   
         istp = nit000 - 1 
         CALL  trj_rea( istp, 1 ) 
         !--------------------------------------------------------------------
         ! Initialize the tangent variables: dy^* = W dy
         !--------------------------------------------------------------------

         un_tl  (:,:,:) = zun_tlin   (:,:,:)
         vn_tl  (:,:,:) = zvn_tlin   (:,:,:)
         tn_tl  (:,:,:) = ztn_tlin   (:,:,:)
         sn_tl  (:,:,:) = zsn_tlin   (:,:,:)
         sshn_tl(  :,:) = zsshn_tlin (  :,:)

#if defined key_pomme_r025
         IF ( (jpert == 5) .OR. (jpert == jpertmax) ) THEN
            !DO ji = 1, jpi
            !   DO jj = 1, jpj
            !      sshn_tl(ji,jj) = cos( (2.*rpi)*(FLOAT(ji)-0.5)/FLOAT(jpi) + rpi/2. ) &
            !                     * sin( (2.*rpi)*(FLOAT(jj)-0.5)/FLOAT(jpj) ) / 100.
            !   END DO
            !END DO
            
            DO ji = 1, jpi
               DO jj = 1, jpj
                  zsshn_tlin(ji,jj) = exp( -((float(ji)-float(jpi)/2.)/(float(jpi)/5.))**2 &
                                           -((float(jj)-float(jpj)/2.)/(float(jpj)/5.))**2 ) / 100. &
                                      * tmask(ji,jj,1)
               END DO
            END DO
            sshn_tl(:,:) = zsshn_tlin(:,:) 
         ENDIF
#endif

         !-----------------------------------------------------------------------
         !  Initialization of the dynamics and tracer fields for the tangent
         !-----------------------------------------------------------------------

         CALL istate_init_tan 
         
         DO istp = nit000, nitend, 1
            CALL stp_tan( istp )
         END DO


         zun_tlout  ( :,:,:) = un_tl   (:,:,:)
         zvn_tlout  ( :,:,:) = vn_tl   (:,:,:)
         ztn_tlout  ( :,:,:) = tn_tl   (:,:,:)
         zsn_tlout  ( :,:,:) = sn_tl   (:,:,:)
         zsshn_tlout(   :,:) = sshn_tl (  :,:)

#if defined key_obc
         zub_tlout  (:,:,:) = ub_tl  (:,:,:)
         zvb_tlout  (:,:,:) = vb_tl  (:,:,:)
         ztb_tlout  (:,:,:) = tb_tl  (:,:,:)
         zsb_tlout  (:,:,:) = sb_tl  (:,:,:)
         zsshb_tlout(:,:)   = sshb_tl(:,:)
#endif

         !--------------------------------------------------------------------
         ! Initialize the adjoint variables: dy^* = W dy
         !--------------------------------------------------------------------
         
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zun_adin(ji,jj,jk) = zun_tlout(ji,jj,jk) &
                     &               * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,jk) &
                     &               * umask(ji,jj,jk) * wesp_u
#if defined key_obc
                  zub_adin(ji,jj,jk) = zub_tlout(ji,jj,jk) &
                     &               * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,jk) &
                     &               * umask(ji,jj,jk) * wesp_u
#endif
               END DO
            END DO
         END DO
         
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zvn_adin(ji,jj,jk) = zvn_tlout(ji,jj,jk) &
                     &               * e1v(ji,jj) * e2v(ji,jj) * fse3v(ji,jj,jk) &
                     &               * vmask(ji,jj,jk) * wesp_u
#if defined key_obc
                  zvb_adin(ji,jj,jk) = zvb_tlout(ji,jj,jk) &
                     &               * e1v(ji,jj) * e2v(ji,jj) * fse3v(ji,jj,jk) &
                     &               * vmask(ji,jj,jk) * wesp_u
#endif
               END DO
            END DO
         END DO

         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  ztn_adin(ji,jj,jk) = ztn_tlout(ji,jj,jk) &
                     &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                     &               * tmask(ji,jj,jk) * wesp_t(jk)
#if defined key_obc
                  ztb_adin(ji,jj,jk) = ztb_tlout(ji,jj,jk) &
                     &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                     &               * tmask(ji,jj,jk) * wesp_t(jk)
#endif
               END DO
            END DO
         END DO
         
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zsn_adin(ji,jj,jk) = zsn_tlout(ji,jj,jk) &
                     &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                     &               * tmask(ji,jj,jk) * wesp_s(jk)
#if defined key_obc
                  zsb_adin(ji,jj,jk) = zsb_tlout(ji,jj,jk) &
                     &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                     &               * tmask(ji,jj,jk) * wesp_s(jk)
#endif
               END DO
            END DO
         END DO
         
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zsshn_adin(ji,jj) = zsshn_tlout(ji,jj) &
                  &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,1) &
                  &               * tmask(ji,jj,1) * wesp_ssh
#if defined key_obc
               zsshb_adin(ji,jj) = zsshb_tlout(ji,jj) &
                  &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,1) &
                  &               * tmask(ji,jj,1) * wesp_ssh
#endif
            END DO
         END DO

         !--------------------------------------------------------------------
         ! Compute the scalar product: ( L dx )^T W dy
         !--------------------------------------------------------------------
         

         zsp1_U    = DOT_PRODUCT( zun_tlout  , zun_adin    )
         zsp1_V    = DOT_PRODUCT( zvn_tlout  , zvn_adin    )
         zsp1_T    = DOT_PRODUCT( ztn_tlout  , ztn_adin    )
         zsp1_S    = DOT_PRODUCT( zsn_tlout  , zsn_adin    )
         zsp1_SSH  = DOT_PRODUCT( zsshn_tlout, zsshn_adin  )
         
         zsp1      = zsp1_U + zsp1_V + zsp1_T + zsp1_S + zsp1_SSH
 
#if defined key_obc
         zsp1_U    = DOT_PRODUCT( zub_tlout  , zub_adin    )
         zsp1_V    = DOT_PRODUCT( zvb_tlout  , zvb_adin    )
         zsp1_T    = DOT_PRODUCT( ztb_tlout  , ztb_adin    )
         zsp1_S    = DOT_PRODUCT( zsb_tlout  , zsb_adin    )
         zsp1_SSH  = DOT_PRODUCT( zsshb_tlout, zsshb_adin  )
         
         zsp1      = zsp1 + ( zsp1_U + zsp1_V + zsp1_T + zsp1_S + zsp1_SSH )
#endif
         !--------------------------------------------------------------------
         ! Call the adjoint routine: dx^* = L^T dy^*
         !--------------------------------------------------------------------
         
         un_ad  (:,:,:) = zun_adin   (:,:,:)
         vn_ad  (:,:,:) = zvn_adin   (:,:,:)
         tn_ad  (:,:,:) = ztn_adin   (:,:,:)
         sn_ad  (:,:,:) = zsn_adin   (:,:,:)
         sshn_ad(  :,:) = zsshn_adin (  :,:)   
         
#if defined key_obc
         ub_ad  (:,:,:) = zub_adin   (:,:,:)
         vb_ad  (:,:,:) = zvb_adin   (:,:,:)
         tb_ad  (:,:,:) = ztb_adin   (:,:,:)
         sb_ad  (:,:,:) = zsb_adin   (:,:,:)
         sshb_ad(:,:)   = zsshb_adin (:,:)   
#endif

         DO istp = nitend, nit000, -1
            CALL stp_adj(istp)
         END DO
         
         CALL istate_init_adj
         
         zun_adout  ( :,:,:) = un_ad   (:,:,:)
         zvn_adout  ( :,:,:) = vn_ad   (:,:,:)
         ztn_adout  ( :,:,:) = tn_ad   (:,:,:)
         zsn_adout  ( :,:,:) = sn_ad   (:,:,:)
         zsshn_adout(   :,:) = sshn_ad (  :,:)
         
         zsp2_U    = DOT_PRODUCT( zun_tlin  , zun_adout )
         zsp2_V    = DOT_PRODUCT( zvn_tlin  , zvn_adout )
         zsp2_T    = DOT_PRODUCT( ztn_tlin  , ztn_adout )
         zsp2_S    = DOT_PRODUCT( zsn_tlin  , zsn_adout )
         zsp2_SSH  = DOT_PRODUCT( zsshn_tlin, zsshn_adout )
         
         zsp2      = zsp2_U + zsp2_V + zsp2_T + zsp2_S + zsp2_SSH
         
         ! 14 char:'12345678901234'
         SELECT CASE (jpert)
         CASE(1)
            cl_name = 'step_adj     U'
         CASE(2)
            cl_name = 'step_adj     V'
         CASE(3)
            cl_name = 'step_adj     T'
         CASE(4)
            cl_name = 'step_adj     S'
         CASE(5)
            cl_name = 'step_adj   SSH'
         CASE(jpertmax)
            cl_name = 'step_adj      '
         END SELECT
         CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )
         endif
      END DO

      DEALLOCATE(                                 &
         & zun_tlin   , zvn_tlin   , ztn_tlin   , &
         & zsn_tlin   , zsshn_tlin , zun_tlout  , &
         & zvn_tlout  , ztn_tlout  , zsn_tlout  , &
         & zsshn_tlout, zun_adin   , zvn_adin   , &
         & ztn_adin   , zsn_adin   , zsshn_adin , &
         & zun_adout  , zvn_adout  , ztn_adout  , &
         & zsn_adout  , zsshn_adout, z2r        , &
         & z3r                                    &
         & )

   END SUBROUTINE stp_adj_tst
#if defined key_tst_tlm
   SUBROUTINE stp_tlm_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE example_tl_tst ***
      !!
      !! ** Purpose : Test the tangent linear routine.
      !!
      !! ** Method  : Verify the tangent with Taylor expansion 
      !!           
      !!                 M(x+hdx) = M(x) + L(hdx) + O(h^2)
      !!
      !!              where  L   = tangent routine
      !!                     M   = direct routine
      !!                     dx  = input perturbation (random field)
      !!                     h   = ration on perturbation 
      !! 
      !!    In the tangent test we verify that:
      !!                M(x+h*dx) - M(x)
      !!        g(h) = ------------------ --->  1    as  h ---> 0
      !!                    L(h*dx)
      !!    and
      !!                g(h) - 1
      !!        f(h) = ----------         --->  k (costant) as  h ---> 0
      !!                    p
      !!                      
      !! History :
      !!        ! 09-06 (F. Vigilant)
      !!-----------------------------------------------------------------------
      !! * Modules used
      USE opatam_tst_ini, ONLY: &
       & tlm_namrd
      USE sbcssr_tam, ONLY: &
         & qrp_tl,erp_tl
      USE sbcfwb_tam, ONLY: &
         & a_fwb_tl           ! for before year.
      USE step_c1d            ! Time stepping loop for the 1D configuration
      USE tamtrj              ! writing out state trajectory  
      USE lib_mpp,    ONLY: & ! distributed memory computing
        & lk_mpp,           &
        & mpp_max 
      USE c1d,        ONLY: & ! 1D configuration
        & lk_c1d
      USE par_tlm,    ONLY: &
        & tlm_bch,          &
        & cur_loop,         &
        & h_ratio
      USE istate_mod
      USE wzvmod             !  vertical velocity
      USE sbc_oce           , ONLY: & ! ocean dynamics and tracers variables
        & emp,              &
        & emps
      USE tamctl,         ONLY: & ! Control parameters
       & numtan, numtan_sc
      USE step
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt               ! Output unit
 
      !! * Local declarations
      INTEGER ::             &
         & ji,               &  ! dummy loop indices
         & jj,               &        
         & jk,               &
         & istp,             &
         & pgamma,           &  ! control the number of pertubation loops
         & igamma,           &
         & nit_loop
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d             ! 2D seed for the random number generator
      REAL(KIND=wp) ::       &
         & zsp1, zsp2, zsp3, zzsp,                   &  ! scalar product
         & zsp1_U, zsp1_V, zsp1_T, zsp1_S, zsp1_SSH, &   
         & zsp2_U, zsp2_V, zsp2_T, zsp2_S, zsp2_SSH, &
         & zsp3_U, zsp3_V, zsp3_T, zsp3_S, zsp3_SSH, &
         & zzsp_U, zzsp_V, zzsp_T, zzsp_S, zzsp_SSH, &
         & gamma,                                    &
         & zgsp1, zgsp2, zgsp3, zgsp4, zgsp5,     &
         & zgsp6, zgsp7     
      REAL(KIND=wp) ::       &  
         & zgsp1_U, zgsp1_V, zgsp1_T, zgsp1_S, zgsp1_SSH,     &
         & zgsp4_U, zgsp4_V, zgsp4_T, zgsp4_S, zgsp4_SSH,      &
         & zgsp5_U, zgsp5_V, zgsp5_T, zgsp5_S, zgsp5_SSH     
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & ta_tmp   ,        &
         & sa_tmp   ,        &
         & zun_tlin ,        & ! Pertubation input
         & zvn_tlin ,        & ! Pertubation input
         & ztn_tlin ,        & ! Pertubation input
         & zsn_tlin ,        & ! Pertubation input
         & zun_out  ,        & ! Direct output
         & zvn_out  ,        & ! Direct output
         & ztn_out  ,        & ! Direct output
         & zsn_out  ,        & ! Direct output
         & zun_wop  ,        & ! Direct without pertubation output
         & zvn_wop  ,        & ! Direct without pertubation output
         & ztn_wop  ,        & ! Direct without pertubation output
         & zsn_wop  ,        & ! Direct without pertubation output
         & z3r                  ! 3D random field  
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & emp_tmp,              &
         & emps_tmp,             &
         & sshn_tmp ,             &
         & zsshn_tlin  ,     & ! Pertubation input
         & zsshn_out   ,     & ! Direct output
         & zsshn_wop    ,     & ! Direct without pertubation output
         & z2r                 ! 2D random field 
 
      CHARACTER(LEN=14)   :: cl_name
      CHARACTER (LEN=128) :: file_out, file_wop, file_wop2, file_xdx
      CHARACTER (LEN=90)  :: FMT

      REAL(KIND=wp), DIMENSION(100):: &
         & zscu   , zscv   , zsct   , zscs   , zscssh, &
         & zscerru, zscerrv, zscerrt, zscerrs, zscerrssh
      INTEGER, DIMENSION(100)::       &
         & iiposu, iiposv, iipost, iiposs, iiposssh, &
         & ijposu, ijposv, ijpost, ijposs, ijposssh, &
         & ikposu, ikposv, ikpost, ikposs
      INTEGER::                &
         & ii,                 &
         & isamp=40,           &
         & jsamp=40,           &
         & ksamp=10,           &
         & numsctlm
     REAL(KIND=wp), DIMENSION(jpi,jpj,jpk) :: &
         & zerru, zerrv, zerrt, zerrs            ! Pertubation input
      REAL(KIND=wp), DIMENSION(jpi,jpj)    :: &
         & zerrssh                               ! Pertubation input
      INTEGER ::               &
         & jdir, jdirmax=6                          ! Perturbation direction (U, V, T, S, SSH, all)

      ! Allocate memory

      ALLOCATE( &
         & zun_tlin(   jpi,jpj,jpk),     &
         & zvn_tlin(   jpi,jpj,jpk),     &
         & ztn_tlin(   jpi,jpj,jpk),     &
         & zsn_tlin(   jpi,jpj,jpk),     &
         & zsshn_tlin( jpi,jpj    ),     &
         & zun_out(    jpi,jpj,jpk),     &
         & zvn_out(    jpi,jpj,jpk),     &
         & ztn_out(    jpi,jpj,jpk),     &
         & zsn_out(    jpi,jpj,jpk),     &
         & zsshn_out(  jpi,jpj    ),     &
         & zun_wop(    jpi,jpj,jpk),     &
         & zvn_wop(    jpi,jpj,jpk),     &
         & ztn_wop(    jpi,jpj,jpk),     &
         & zsn_wop(    jpi,jpj,jpk),     &
         & zsshn_wop(  jpi,jpj    ),     &
         & z2r(        jpi,jpj    ),     &
         & z3r(        jpi,jpj,jpk),     &
         & emp_tmp(    jpi,jpj    ),     &
         & emps_tmp(   jpi,jpj    ),     &
         & ta_tmp(     jpi,jpj,jpk),     &
         & sa_tmp(     jpi,jpj,jpk),     &
         & sshn_tmp(    jpi,jpj   )      &  
         & )

      !--------------------------------------------------------------------
      ! Reset the scalar
      !--------------------------------------------------------------------
      zscu   (:) = 0.0_wp
      zscv   (:) = 0.0_wp
      zsct   (:) = 0.0_wp
      zscs   (:) = 0.0_wp
      zscssh (:) = 0.0_wp

      iiposu   (:) = 0
      iiposv   (:) = 0
      iipost   (:) = 0
      iiposs   (:) = 0
      iiposssh (:) = 0
      ijposu   (:) = 0
      ijposv   (:) = 0
      ijpost   (:) = 0
      ijposs   (:) = 0
      ijposssh (:) = 0
      ikposu   (:) = 0
      ikposv   (:) = 0
      ikpost   (:) = 0
      ikposs   (:) = 0

      zscerru   (:) = 0.0_wp
      zscerrv   (:) = 0.0_wp
      zscerrt   (:) = 0.0_wp
      zscerrs   (:) = 0.0_wp
      zscerrssh (:) = 0.0_wp

      zerru   (:,:,:) = 0.0_wp
      zerrv   (:,:,:) = 0.0_wp
      zerrt   (:,:,:) = 0.0_wp
      zerrs   (:,:,:) = 0.0_wp
      zerrssh (:,:) = 0.0_wp

      !--------------------------------------------------------------------
      ! Reset the tangent variables
      !--------------------------------------------------------------------
      zun_tlin   ( :,:,:) = 0.0_wp
      zvn_tlin   ( :,:,:) = 0.0_wp
      ztn_tlin   ( :,:,:) = 0.0_wp
      zsn_tlin   ( :,:,:) = 0.0_wp
      zsshn_tlin (   :,:) = 0.0_wp
      z3r        ( :,:,:) = 0.0_wp
      z2r        (   :,:) = 0.0_wp

      !--------------------------------------------------------------------
      ! Output filename Xn=F(X0)
      !--------------------------------------------------------------------
      CALL tlm_namrd
      gamma = h_ratio
      file_wop='trj_wop_step'
      file_xdx='trj_xdx_step'
      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------
      jdir = jdirmax
       IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
          IF (( jdir .EQ. 1 ) .OR. ( jdir .EQ. jdirmax ) ) THEN
             CALL grid_rd_sd( 596035, z3r,  'U', 0.0_wp, stdu)
             DO jk = 1, jpk
                DO jj = nldj, nlej
                   DO ji = nldi, nlei
                      zun_tlin(ji,jj,jk) = z3r(ji,jj,jk)
                   END DO
                END DO
             END DO 

          ENDIF
          IF (( jdir .EQ. 2 ) .OR. ( jdir .EQ. jdirmax ) ) THEN
             CALL grid_rd_sd( 371836, z3r,  'V', 0.0_wp, stdv)
             DO jk = 1, jpk
                DO jj = nldj, nlej
                   DO ji = nldi, nlei
                      zvn_tlin(ji,jj,jk) = z3r(ji,jj,jk)
                   END DO
                END DO
             END DO

          ENDIF
          IF (( jdir .EQ. 3 ) .OR. ( jdir .EQ. jdirmax ) ) THEN
          CALL grid_rd_sd( 284035, z3r,  'T', 0.0_wp, stdt)
             DO jk = 1, jpk
                DO jj = nldj, nlej
                   DO ji = nldi, nlei
                      ztn_tlin(ji,jj,jk) = z3r(ji,jj,jk)
                   END DO
                END DO
             END DO

          ENDIF
          IF (( jdir .EQ. 4 ) .OR. ( jdir .EQ. jdirmax ) ) THEN
             CALL grid_rd_sd( 471426, z3r,  'T', 0.0_wp, stds)
             DO jk = 1, jpk
                DO jj = nldj, nlej
                   DO ji = nldi, nlei
                      zsn_tlin(ji,jj,jk) = z3r(ji,jj,jk)
                   END DO
                END DO
             END DO  

          ENDIF
          IF (( jdir .EQ. 5 ) .OR. ( jdir .EQ. jdirmax ) ) THEN
             CALL grid_rd_sd( 712035, z2r,'T', 0.0_wp, stdssh)
             DO jj = nldj, nlej
                DO ji = nldi, nlei
                   zsshn_tlin(ji,jj) = z2r(ji,jj)
                END DO
             END DO

          ENDIF
     ENDIF 
     IF ( tlm_bch /= 2 )  CALL istate_p 

         !--------------------------------------------------------------------
         !  Open File to save scalar result
         !--------------------------------------------------------------------

      ! check that all process are still there... If some process have an error,
      ! they will never enter in step and other processes will wait until the end of the cpu time!
     IF( lk_mpp )   CALL mpp_max( nstop )

        istp = nit000
        IF ( tlm_bch /= 2 )  THEN
        IF( lk_c1d ) THEN                 ! 1D configuration (no AGRIF zoom)
           DO WHILE ( istp <= nitend .AND. nstop == 0 )
              CALL stp_c1d( istp )
              istp = istp + 1
           END DO
        ENDIF
        ENDIF

        IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
           zun_tlin(:,:,:) = gamma * zun_tlin(:,:,:)
           un(:,:,:)       = un(:,:,:) + zun_tlin(:,:,:)

           zvn_tlin(:,:,:) = gamma * zvn_tlin(:,:,:)
           vn(:,:,:)        = vn(:,:,:) + zvn_tlin(:,:,:)

           ztn_tlin(:,:,:) = gamma * ztn_tlin(:,:,:)
           tn(:,:,:)        = tn(:,:,:) + ztn_tlin(:,:,:)

           zsn_tlin(:,:,:) = gamma * zsn_tlin(:,:,:)
           sn(:,:,:)        = sn(:,:,:) + zsn_tlin(:,:,:)

           zsshn_tlin(:,:)   = gamma * zsshn_tlin(:,:)
           sshn (:,:)        = sshn (:,:) + zsshn_tlin(:,:)
        ENDIF 
 
        IF( .NOT. lk_vvl ) CALL wzv(nit000) 

        !--------------------------------------------------------------------
        !  Compute the direct model F(X0,t=n) = Xn
        !--------------------------------------------------------------------
        IF ( tlm_bch /= 2 )  THEN
        DO istp = nit000, nitend, 1
           CALL stp( istp )
        END DO
        ENDIF
        IF ( tlm_bch == 0 ) CALL trj_wri_spl(file_wop)
        IF ( tlm_bch == 1 ) CALL trj_wri_spl(file_xdx)
        !--------------------------------------------------------------------
        !  Compute the Tangent 
        !--------------------------------------------------------------------
        IF ( tlm_bch == 2 ) THEN        

           !--------------------------------------------------------------------
           ! Initialize the tangent variables: dy^* = W dy  
           !--------------------------------------------------------------------
           qrp_tl = 0.0_wp
#if defined key_tradmp
           strdmp_tl = 0.0_wp
           ttrdmp_tl = 0.0_wp
#endif
           a_fwb_tl = 0.0_wp

           CALL  trj_rea( nit000-1, 1 ) 

           un_tl  (:,:,:) = zun_tlin   (:,:,:)
           vn_tl  (:,:,:) = zvn_tlin   (:,:,:)
           tn_tl  (:,:,:) = ztn_tlin   (:,:,:)
           sn_tl  (:,:,:) = zsn_tlin   (:,:,:)
           sshn_tl(  :,:) = zsshn_tlin (  :,:)
 

           !-----------------------------------------------------------------------
           !  Initialization of the dynamics and tracer fields for the tangent
           !-----------------------------------------------------------------------

           CALL istate_init_tan 
 
           DO istp = nit000, nitend, 1
              CALL stp_tan( istp )
              !CALL stp_tan_cpd( istp )
           END DO
       
           !--------------------------------------------------------------------
           ! Compute the scalar product: ( L(t0,tn) gamma dx0 ) )
           !--------------------------------------------------------------------

           zsp2_U    = DOT_PRODUCT( un_tl  , un_tl    )
           zsp2_V    = DOT_PRODUCT( vn_tl  , vn_tl    )
           zsp2_T    = DOT_PRODUCT( tn_tl  , tn_tl    )
           zsp2_S    = DOT_PRODUCT( sn_tl  , sn_tl    )
           zsp2_SSH  = DOT_PRODUCT( sshn_tl, sshn_tl  )

           zsp2      = zsp2_U + zsp2_V + zsp2_T + zsp2_S + zsp2_SSH
  
           !--------------------------------------------------------------------
           !  Storing data
           !--------------------------------------------------------------------
           CALL trj_rd_spl(file_wop)
           zun_wop   (:,:,:) = un   (:,:,:)
           zvn_wop   (:,:,:) = vn   (:,:,:)
           ztn_wop   (:,:,:) = tn   (:,:,:)
           zsn_wop   (:,:,:) = sn   (:,:,:)
           zsshn_wop (:,:  ) = sshn (:,:  )  
           CALL trj_rd_spl(file_xdx) 
           zun_out   (:,:,:) = un   (:,:,:)
           zvn_out   (:,:,:) = vn   (:,:,:)
           ztn_out   (:,:,:) = tn   (:,:,:)
           zsn_out   (:,:,:) = sn   (:,:,:)
           zsshn_out (:,:  ) = sshn (:,:  )
           !--------------------------------------------------------------------
           ! Compute the Linearization Error 
           ! Nn = M( X0+gamma.dX0, t0,tn) - M(X0, t0,tn)
           ! and 
           ! Compute the Linearization Error 
           ! En = Nn -TL(gamma.dX0, t0,tn)
           !--------------------------------------------------------------------
           ! Warning: Here we re-use local variables z()_out and z()_wop
           ii=0
           DO jk = 1, jpk
              DO jj = 1, jpj
                 DO ji = 1, jpi
                    zun_out   (ji,jj,jk) = zun_out    (ji,jj,jk) - zun_wop  (ji,jj,jk)
                    zun_wop   (ji,jj,jk) = zun_out    (ji,jj,jk) - un_tl    (ji,jj,jk)
                    IF (  un_tl(ji,jj,jk) .NE. 0.0_wp ) &
                       & zerru(ji,jj,jk) = zun_out(ji,jj,jk)/un_tl(ji,jj,jk)
                       IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                       &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                       &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                          ii = ii+1
                          iiposu(ii) = ji
                          ijposu(ii) = jj
                          ikposu(ii) = jk
                          IF ( INT(umask(ji,jj,jk)) .NE. 0)  THEN
                             zscu (ii) =  zun_out(ji,jj,jk)
                             zscerru (ii) = ( zerru(ji,jj,jk) - 1.0_wp ) / gamma
                          ENDIF
                    ENDIF
                 END DO
              END DO
           END DO
           ii=0      
           DO jk = 1, jpk
              DO jj = 1, jpj
                 DO ji = 1, jpi
                    zvn_out   (ji,jj,jk) = zvn_out    (ji,jj,jk) - zvn_wop  (ji,jj,jk)
                    zvn_wop   (ji,jj,jk) = zvn_out    (ji,jj,jk) - vn_tl    (ji,jj,jk)
                    IF (  vn_tl(ji,jj,jk) .NE. 0.0_wp ) &
                    & zerrv(ji,jj,jk) = zvn_out(ji,jj,jk)/vn_tl(ji,jj,jk)
                       IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                       &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                       &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                          ii = ii+1
                          iiposv(ii) = ji
                          ijposv(ii) = jj
                          ikposv(ii) = jk
                          IF ( INT(vmask(ji,jj,jk)) .NE. 0)  THEN
                             zscv (ii) =  zvn_out    (ji,jj,jk)
                             zscerrv (ii) =  ( zerrv(ji,jj,jk) - 1.0_wp ) / gamma
                          ENDIF
                       ENDIF
                 END DO
              END DO
           END DO
           ii=0
           DO jk = 1, jpk
              DO jj = 1, jpj
                 DO ji = 1, jpi
                    ztn_out   (ji,jj,jk) = ztn_out   (ji,jj,jk) - ztn_wop   (ji,jj,jk)
                    ztn_wop   (ji,jj,jk) = ztn_out   (ji,jj,jk) - tn_tl     (ji,jj,jk)
                    IF (  tn_tl(ji,jj,jk) .NE. 0.0_wp ) &
                    & zerrt(ji,jj,jk) = ztn_out(ji,jj,jk)/tn_tl(ji,jj,jk)
                       IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                       &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                       &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                          ii = ii+1
                          iipost(ii) = ji
                          ijpost(ii) = jj
                          ikpost(ii) = jk
                          IF ( INT(tmask(ji,jj,jk)) .NE. 0)  THEN
                             zsct (ii) =  ztn_out    (ji,jj,jk)
                             zscerrt (ii) =  ( zerrt(ji,jj,jk) - 1.0_wp ) / gamma
                          ENDIF
                       ENDIF
                 END DO
              END DO
           END DO
           ii=0
           DO jk = 1, jpk
              DO jj = 1, jpj
                 DO ji = 1, jpi
                    zsn_out   (ji,jj,jk) = zsn_out   (ji,jj,jk) - zsn_wop   (ji,jj,jk)
                    zsn_wop   (ji,jj,jk) = zsn_out   (ji,jj,jk) - sn_tl     (ji,jj,jk)
                    IF (  sn_tl(ji,jj,jk) .NE. 0.0_wp ) &
                    & zerrs(ji,jj,jk) = zsn_out(ji,jj,jk)/sn_tl(ji,jj,jk)
                       IF( (MOD(ji, isamp) .EQ. 0) .AND. & 
                       &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                       &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                          ii = ii+1
                          iiposs(ii) = ji
                          ijposs(ii) = jj
                          ikposs(ii) = jk
                          IF ( INT(tmask(ji,jj,jk)) .NE. 0)  THEN
                             zscs (ii) =  zsn_out    (ji,jj,jk)
                             zscerrs (ii) =  ( zerrs(ji,jj,jk) - 1.0_wp ) / gamma
                          ENDIF
                       ENDIF
                 END DO
              END DO
           END DO
           ii=0
           jk=1
           DO jj = 1, jpj
              DO ji = 1, jpi
                 zsshn_out (ji,jj   ) = zsshn_out (ji,jj   ) - zsshn_wop (ji,jj   )
                 zsshn_wop (ji,jj   ) = zsshn_out (ji,jj   ) - sshn_tl   (ji,jj   )
                 IF (  sshn_tl(ji,jj) .NE. 0.0_wp ) &
                 & zerrssh(ji,jj) = zsshn_out(ji,jj)/sshn_tl(ji,jj)
                    IF( (MOD(ji, isamp) .EQ. 0) .AND. & 
                    &   (MOD(jj, jsamp) .EQ. 0) ) THEN
                       ii = ii+1
                       iiposssh(ii) = ji
                       ijposssh(ii) = jj
                       IF ( INT(tmask(ji,jj,jk)) .NE. 0 ) THEN
                          zscssh (ii) =  zsshn_out    (ji,jj)
                          zscerrssh (ii) =  ( zerrssh(ji,jj) - 1.0_wp ) / gamma
                       ENDIF
                    ENDIF
              END DO
           END DO

           zsp1_U    = DOT_PRODUCT( zun_out  , zun_out    )
           zsp1_V    = DOT_PRODUCT( zvn_out  , zvn_out    )
           zsp1_T    = DOT_PRODUCT( ztn_out  , ztn_out    )
           zsp1_S    = DOT_PRODUCT( zsn_out  , zsn_out    )
           zsp1_SSH  = DOT_PRODUCT( zsshn_out, zsshn_out  )

           zsp1      = zsp1_U + zsp1_V + zsp1_T + zsp1_S + zsp1_SSH

           zsp3_U    = DOT_PRODUCT( zun_wop  , zun_wop    )
           zsp3_V    = DOT_PRODUCT( zvn_wop  , zvn_wop    )
           zsp3_T    = DOT_PRODUCT( ztn_wop  , ztn_wop    )
           zsp3_S    = DOT_PRODUCT( zsn_wop  , zsn_wop    )
           zsp3_SSH  = DOT_PRODUCT( zsshn_wop, zsshn_wop  )

           zsp3      = zsp3_U + zsp3_V + zsp3_T + zsp3_S + zsp3_SSH

         !--------------------------------------------------------------------
         ! Print the linearization error En - norme 2
         !--------------------------------------------------------------------
         ! 14 char:'12345678901234'
         cl_name = 'step_tam:En   '
         zzsp     = SQRT(zsp3)
         zzsp_U   = SQRT(zsp3_U)
         zzsp_V   = SQRT(zsp3_V)
         zzsp_T   = SQRT(zsp3_T)
         zzsp_S   = SQRT(zsp3_S)
         zzsp_SSH = SQRT(zsp3_SSH)
         zgsp5    = zzsp
         zgsp5_U=zzsp_U
         zgsp5_V=zzsp_V
         zgsp5_T=zzsp_T
         zgsp5_S=zzsp_S
         zgsp5_SSH=zzsp_SSH
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         CALL prntst_tlm( 'U             ', kumadt, zzsp_U  , h_ratio )
         CALL prntst_tlm( 'V             ', kumadt, zzsp_V  , h_ratio )
         CALL prntst_tlm( 'T             ', kumadt, zzsp_T  , h_ratio )
         CALL prntst_tlm( 'S             ', kumadt, zzsp_S  , h_ratio )
         CALL prntst_tlm( 'SSH           ', kumadt, zzsp_SSH, h_ratio )

         !--------------------------------------------------------------------
         ! Compute TLM norm2
         !--------------------------------------------------------------------
         zzsp     =  SQRT(zsp2)
         zzsp_U   =  SQRT(zsp2_U)
         zzsp_V   =  SQRT(zsp2_V)
         zzsp_T   =  SQRT(zsp2_T)
         zzsp_S   =  SQRT(zsp2_S)
         zzsp_SSH =  SQRT(zsp2_SSH)
         zgsp4    = zzsp
         zgsp4_U=zzsp_U
         zgsp4_V=zzsp_V
         zgsp4_T=zzsp_T
         zgsp4_S=zzsp_S
         zgsp4_SSH=zzsp_SSH
         cl_name = 'step_tam:Ln2  '
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         CALL prntst_tlm( 'U             ', kumadt, zzsp_U  , h_ratio )
         CALL prntst_tlm( 'V             ', kumadt, zzsp_V  , h_ratio )
         CALL prntst_tlm( 'T             ', kumadt, zzsp_T  , h_ratio )
         CALL prntst_tlm( 'S             ', kumadt, zzsp_S  , h_ratio )
         CALL prntst_tlm( 'SSH           ', kumadt, zzsp_SSH, h_ratio )

         !--------------------------------------------------------------------
         ! Print the linearization error Nn - norme 2
         !--------------------------------------------------------------------
         ! 14 char:'12345678901234'
         cl_name = 'step_tam:Nn   '
         zzsp     =  SQRT(zsp1)
         zzsp_U   =  SQRT(zsp1_U)
         zzsp_V   =  SQRT(zsp1_V)
         zzsp_T   =  SQRT(zsp1_T)
         zzsp_S   =  SQRT(zsp1_S)
         zzsp_SSH =  SQRT(zsp1_SSH)
!         zzsp      = zzsp_U + zzsp_V + zzsp_T + zzsp_S + zzsp_SSH
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         CALL prntst_tlm( 'U             ', kumadt, zzsp_U  , h_ratio )
         CALL prntst_tlm( 'V             ', kumadt, zzsp_V  , h_ratio )
         CALL prntst_tlm( 'T             ', kumadt, zzsp_T  , h_ratio )
         CALL prntst_tlm( 'S             ', kumadt, zzsp_S  , h_ratio )
         CALL prntst_tlm( 'SSH           ', kumadt, zzsp_SSH, h_ratio )

         zgsp3    = SQRT( zsp3/zsp2 )
         zgsp7    = zgsp3/gamma
         zgsp1    = zzsp
         zgsp1_U=zzsp_U
         zgsp1_V=zzsp_V
         zgsp1_T=zzsp_T
         zgsp1_S=zzsp_S
         zgsp1_SSH=zzsp_SSH
         zgsp2    = zgsp1 / zgsp4
         zgsp6    = (zgsp2 - 1.0_wp)/gamma

         FMT = "(A8,2X,I4.4,2X,E6.1,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13)"

         WRITE(numtan,FMT) 'step    ', cur_loop, h_ratio, zgsp1, zgsp1_T ,zgsp4_T, zgsp5_T,zgsp1_S,zgsp4_S,zgsp5_S!,
         !--------------------------------------------------------------------
         ! Unitary calculus
         !--------------------------------------------------------------------
         FMT = "(A8,2X,A8,2X,I4.4,2X,E6.1,2X,I4.4,2X,I4.4,2X,I4.4,2X,E20.13,1X)"
         cl_name = 'step '
         IF (lwp) THEN
            DO ii=1, 100, 1
               IF ( zscu(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscu     ', &
                    & cur_loop, h_ratio,  iiposu(ii), ijposu(ii), ikposu(ii), zscu(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscv(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscv     ', &
                    & cur_loop, h_ratio,  iiposv(ii), ijposv(ii), ikposv(ii), zscv(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zsct(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zsct     ', &
                    & cur_loop, h_ratio,  iipost(ii), ijpost(ii), ikpost(ii), zsct(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscs(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscs    ', &
                    & cur_loop, h_ratio,  iiposs(ii), ijposs(ii), ikposs(ii), zscs(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscssh(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscssh  ', &
                    & cur_loop, h_ratio,  iiposssh(ii), ijposssh(ii), ijposssh(ii), zscssh(ii)
            ENDDO

            DO ii=1, 100, 1
               IF ( zscerru(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerru  ', &
                    & cur_loop, h_ratio,  iiposu(ii), ijposu(ii), ikposu(ii), zscerru(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrv(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrv  ', &
                    & cur_loop, h_ratio,  iiposv(ii), ijposv(ii), ikposv(ii), zscerrv(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrt(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrt  ', &
                    & cur_loop, h_ratio,  iipost(ii), ijpost(ii), ikpost(ii), zscerrt(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrs(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrs ', &
                    & cur_loop, h_ratio,  iiposs(ii), ijposs(ii), ikposs(ii), zscerrs(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrssh(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerssh', &
                    & cur_loop, h_ratio,  iiposssh(ii), ijposssh(ii), ijposssh(ii), zscerrssh(ii)
            ENDDO
            ! write separator
            WRITE(numtan_sc,"(A4)") '===='
         ENDIF

      ENDIF

      DEALLOCATE(       &
         & zun_tlin   , &
         & zvn_tlin   , &
         & ztn_tlin   , &
         & zsn_tlin   , &
         & zsshn_tlin , &
         & zun_out    , &
         & zvn_out    , &
         & ztn_out    , &
         & zsn_out    , &
         & zsshn_out  , &
         & zun_wop    , &
         & zvn_wop    , &
         & ztn_wop    , &
         & zsn_wop    , &
         & zsshn_wop  , &
         & z2r        , &
         & z3r          &
         & )

   END SUBROUTINE stp_tlm_tst

      !!----------------------------------------------------------------------

   SUBROUTINE asm_trj_wop_tasa_wri
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE asm_trj_wop_tasa_wri ***
      !!
      !! ** Purpose : Write to file the model state trajectory for use with
      !!              4D-Var.
      !!
      !! ** Method  : 
      !!
      !! ** Action  :
      !!
      !! History : 
      !!        ! 09-07 (F. Vigilant) 
      !!-----------------------------------------------------------------------
      !! *Module udes
      USE iom 
      USE oce           , ONLY: & ! ocean dynamics and tracers variables
      & tn, sn 
      !! * Arguments
      !! * Local declarations
      INTEGER :: &
         & inum                  ! File unit number
      CHARACTER (LEN=50) :: &
         & filename
         ! Define the output file  
         filename='trj_wop_tasa'       
         WRITE(filename, FMT='(A,A)' ) TRIM( filename ), '.nc'
         filename = TRIM( filename )
         CALL iom_open( filename, inum, ldwrt = .TRUE., kiolib = jprstlib)

         ! Output trajectory fields
         CALL iom_rstput( 1, 1, inum, 'tn'   , tn  )
         CALL iom_rstput( 1, 1, inum, 'sn'   , sn  )
         CALL iom_close( inum )
   END SUBROUTINE asm_trj_wop_tasa_wri

   SUBROUTINE asm_trj_wop_tasa_rd
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE asm_trj_wop_tasa_rd ***
      !!
      !! ** Purpose : Read file the model state trajectory for use with
      !!              4D-Var.
      !!
      !! ** Method  : 
      !!
      !! ** Action  :
      !!
      !! History : 
      !!        ! 09-07 (F. Vigilant) 
      !!-----------------------------------------------------------------------
      !! *Module udes
      USE iom                 ! I/O module
      USE oce           , ONLY: & ! ocean dynamics and tracers variables
      & tn, sn 
      !! * Arguments
      !! * Local declarations
      INTEGER :: &
         & inum                  ! File unit number
      CHARACTER (LEN=50) :: &
         & filename
         ! Define the output file  
         filename='trj_wop_tasa'       
         WRITE(filename, FMT='(A,A)' ) TRIM( filename ), '.nc'
         filename = TRIM( filename )
         CALL iom_open( filename, inum)

         ! Output trajectory fields
         CALL iom_get( inum, jpdom_data, 'tn'   , tn, 1  )
         CALL iom_get( inum, jpdom_data, 'sn'   , sn, 1  )
         CALL iom_close( inum )
   END SUBROUTINE asm_trj_wop_tasa_rd


   !!======================================================================
#endif
#endif
END MODULE step_tam
