MODULE tamtst
   !!======================================================================
   !!                       ***  MODULE tamtst ***
   !! NEMOVAR : Testing routine
   !!======================================================================

   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !! * Modules used   
   USE par_kind, ONLY : &  ! Precision variables
      & wp
   USE par_oce, ONLY : &   ! Ocean space and time domain variables
      & jpi, jpj, jpk
   USE dom_oce, ONLY : &   ! Processor area number
      & narea
   USE dynspg_oce    , ONLY: &
      &  lk_dynspg_rl  ! Rigid-lid flag
   USE trabbc        , ONLY: &! bottom boundary condition
      & lk_trabbc
   USE in_out_manager      ! Input/output
   USE iom                 ! netCDF output/input
   USE paresp              ! Weights for energy-type scalar product
   USE tamctl              ! TAM control
   USE sbcmod_tam          ! Tangent/adjoint of surface BCs
   USE eosbn2_tam          ! Tangent/adjoint of eq. of state, Brunt-Vaisala
   USE trasbc_tam          ! Tangent/adjoint of surface BCs application
   USE traqsr_tam          ! Tangent/adjoint of penetrative solar radiation
   USE trabbc_tam          ! Tangent/adjoint of bottom heat flux
   USE tradmp_tam          ! Tangent/adjoint of internal damping trends
   USE traadv_tam          ! Tangent/adjoint of horizontal/vertical advection
   USE cla_tam             ! Tangent/adjoint of cross land advection
   USE cla_div_tam         ! Tangent/adjoint of cross land advection
   USE traldf_tam          ! Tangent/adjoint of lateral mixing
   USE trazdf_tam          ! Tangent/adjoint of vertical diffusion
   USE tranxt_tam          ! Tangent/adjoint of tracers at next time step
   USE zpshde_tam          ! Tangent/adjoint of horiz. derivs. for partial steps
   USE divcur_tam          ! Tangent/adjoint of horiz. div. and rel. vorticity
   USE dynadv_tam          ! Tangent/adjoint of horizontal/vertical advection
   USE dynhpg_tam          ! Tangent/adjoint of horiz. pressure gradient
   USE dynkeg_tam          ! Tangent/adjoint of kinetic energy gradient
   USE dynldf_tam          ! Tangent/adjoint of lateral mixing
   USE dynnxt_tam          ! Tangent/adjoint of dynamics at next time step
   USE dynspg_tam          ! Tangent/adjoint of surface pressure gradient
   USE dynvor_tam          ! Tangent/adjoint of relative and planetary vorticity
   USE dynzdf_tam          ! Tangent/adjoint of vertical diffusion
   USE wzvmod_tam          ! Tangent/adjoint of vertical velocity
   USE step_tam            ! manager of the adjoint ocean time stepping
   USE trj_tam             ! reference trajectory
   USE istate_tam
   USE solsor_tam
#if defined key_obc
   USE obcdyn_tam
#endif

   USE step_tam_cpd        ! Temporary

   IMPLICIT NONE

   !! * Routine accessibility
   PRIVATE

   PUBLIC &
      & tsttam, &          !: Scalar product test of the adjoint routines
      & numadt             !: File unit number for adjoint test output

   !! * Module variables
   INTEGER :: &
      & numadt             !: File unit number for adjoint test output

CONTAINS

   SUBROUTINE tsttam
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE tsttam  ***
      !!
      !! ** Purpose : Apply various tests (linearization, adjoint)
      !!              on the NEMOTAM code.
      !!
      !! ** Method  : 
      !!
      !! ** Action  :
      !!                   
      !! History : 
      !!        ! 07-09 (K. Mogensen) NEMOVAR version based on balubt.F
      !!        ! 07-11 (A. Weaver) Add adjoint tests
      !!        ! 09-03 (I. Mirouze) Filter call depends on type
      !!        ! 09-08 (F. Vigilant) Split TAM and Var and add tangent tests
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments

      !! * Local declarations
      CHARACTER (LEN=128) :: file_out

      ! Open adjoint test output unit

      IF (lwp) THEN

         WRITE(file_out,FMT="('adjoint_test.output_',I4.4)") &
            &   narea-1
         CALL ctlopn( numadt, file_out, 'UNKNOWN', 'FORMATTED',   &
            &         'SEQUENTIAL', 1, numadt, .FALSE., 1 )

         WRITE(numout,*) ' tstopt: Start testing adjoint operators ...'
         WRITE(numout,*) ' ------'

         WRITE(numout,*)
         WRITE(numout,990) file_out
990      FORMAT('          Output in file = ',A20)
         WRITE(numout,*)

         WRITE(numadt,*)
         WRITE(numadt,997) 
         WRITE(numadt,998) 
         WRITE(numadt,999) 
997      FORMAT('  Routine (L)',2X,' ( L * dx )^T W dy ',2X, &
            &   '     dx^T L^T W dy    ',2X,'Rel.',2X,      &
            &   'Mach.',2X,'Status')
998      FORMAT('             ',2X,'                   ',2X, &
            &   '                      ',2X,'Err.',2x,      &
            &   'Eps. ',2X,'      ')
999      FORMAT('  -----------',2X,'-------------------',2X, &
            &   '----------------------',2X,'----',2X,      &
            &   '-----',2X,'------')
         CALL FLUSH(numout)
         CALL FLUSH(numadt)

      ENDIF

      ! Initialize energy weights

      CALL par_esp

#if defined key_tam

      IF ( ln_tst_nemotam ) THEN

         IF ( ln_tst_cpd_tam ) THEN
            ! -----------------------------------------------------
            ! 1) Test the adjoint of the components of M (NEMOTAM)
            ! -----------------------------------------------------
            ! *** initialize the reference trajectory
            ! ------------
            CALL  trj_rea( nit000 - 1, 1 ) 

            ! *** Red-Black SOR solver
            ! ------------
            CALL sol_sor_adj_tst( numadt )

            IF (lwp) WRITE(numadt,*)

            ! *** Surface boundary conditions
            ! ------------
            CALL sbc_adj_tst( numadt )        ! surface boundary conditions

            IF (lwp) WRITE(numadt,*)

            ! *** Vertical physics
            ! --------------------

            CALL bn2_adj_tst( numadt )        ! Brunt-Vaisala frequency

            IF (lwp) WRITE(numadt,*)

            CALL div_cla_adj_tst( numadt )    ! Cross land advection (update hor. divergence)

            IF (lwp) WRITE(numadt,*)

            CALL div_cur_adj_tst( numadt )    ! Horizontal divergence and relative vorticity

            IF (lwp) WRITE(numadt,*)

            CALL dyn_adv_adj_tst( numadt )    ! Advection (vector or flux form)

            IF (lwp) WRITE(numadt,*)

            CALL dyn_vor_adj_tst( numadt )    ! Vorticity term including Coriolis

            IF (lwp) WRITE(numadt,*)

            CALL dyn_ldf_adj_tst( numadt )    ! Lateral mixing

            IF (lwp) WRITE(numadt,*)

            CALL dyn_zdf_adj_tst( numadt )    ! Vertical diffusion

            IF (lwp) WRITE(numadt,*)

            CALL dyn_spg_adj_tst( numadt )    ! Surface pressure gradient

            IF (lwp) WRITE(numadt,*)

#if defined key_obc
            CALL obc_dyn_adj_tst( numadt )    ! OBC

            IF (lwp) WRITE(numadt,*)
#endif

            CALL dyn_hpg_adj_tst( numadt )    ! Horizontal pressure gradient

            IF (lwp) WRITE(numadt,*)

            CALL dyn_nxt_adj_tst( numadt )    ! Lateral velocity at next time step

            IF (lwp) WRITE(numadt,*)

            CALL wzv_adj_tst( numadt )        ! Vertical velocity

            IF (lwp) WRITE(numadt,*)

            ! *** Tracers
            ! -----------

            CALL tra_sbc_adj_tst( numadt )    ! Surface boundary condition

            IF (lwp) WRITE(numadt,*)

            CALL tra_qsr_adj_tst( numadt )    ! Penetrative solar radiation

            IF (lwp) WRITE(numadt,*)

            IF( lk_trabbc      )  &
                & CALL tra_bbc_adj_tst( numadt )    ! Bottom heat flux

            IF (lwp) WRITE(numadt,*)

            IF( lk_tradmp      )  &
                & CALL tra_dmp_adj_tst( numadt )    ! Internal damping trends

            IF (lwp) WRITE(numadt,*)

            CALL tra_adv_adj_tst( numadt )    ! Horizontal and vertical advection

            IF (lwp) WRITE(numadt,*)

            CALL tra_cla_adj_tst( numadt )    ! Cross land advection (update hor. advection)

            IF (lwp) WRITE(numadt,*)

            CALL tra_ldf_adj_tst( numadt )    ! Lateral mixing

            IF (lwp) WRITE(numadt,*)

            CALL tra_zdf_adj_tst( numadt )    ! Vertical mixing

            IF (lwp) WRITE(numadt,*)

            CALL tra_nxt_adj_tst( numadt )    ! Tracer fields at next time step

            IF (lwp) WRITE(numadt,*)

            CALL eos_adj_tst( numadt )        ! In situ density

            IF (lwp) WRITE(numadt,*)

            CALL zps_hde_adj_tst( numadt )    ! Partial steps: horiz. grad. at bottom level

            IF (lwp) WRITE(numadt,*)

            CALL istate_init_adj_tst( numadt )

            IF (lwp) WRITE(numadt,*)

            CALL flush(numout)

            IF (lwp) THEN
               WRITE(numout,*)
               WRITE(numout,*) ' tstopt: Finished testing standalone operators'
               WRITE(numout,*) ' ------'
               WRITE(numout,*)
            ENDIF
         ENDIF

         ! -----------------------------------------------------
         ! 2) Test the adjoint of of M (NEMOTAM)
         ! -----------------------------------------------------
         IF ( ln_tst_stp_tam ) THEN

            CALL stp_adj_tst( numadt )        ! Time-stepping

            IF (lwp) WRITE(numadt,*)

            CALL flush(numout)

         ENDIF
#if defined key_tst_tlm
         ! -----------------------------------------------------
         ! 3)  Test the Tangent accuracy
         ! ----------------------
         IF ( ln_tst_tan ) THEN     

            IF (ln_tst_tan_cpd) THEN

               CALL dyn_hpg_tlm_tst( numadt )

               CALL dyn_spg_tlm_tst( numadt )

               CALL sol_sor_tlm_tst( numadt )

               CALL zps_hde_tlm_tst( numadt )

               CALL tra_sbc_tlm_tst( numadt )

               CALL dyn_adv_tlm_tst( numadt )

               CALL eos_tlm_tst( numadt )

               CALL bn2_tlm_tst( numadt )

               CALL tra_zdf_tlm_tst( numadt )

               CALL tra_adv_tlm_tst( numadt )       

               CALL tra_ldf_tlm_tst( numadt ) 
       
            ELSE

               CALL stp_tlm_tst( numadt )

               IF (lwp) WRITE(numadt,*)

               CALL flush(numout)
 
            ENDIF

         ENDIF
#endif
      ENDIF
#endif
      ! Close output file

      IF (lwp) CLOSE(numadt)

      IF (lwp) THEN
         WRITE(numout,*)
         WRITE(numout,*) ' tsttam: Finished testing operators'
         WRITE(numout,*) ' ------'
         WRITE(numout,*)
      ENDIF
      CALL flush(numout)
   END SUBROUTINE tsttam

END MODULE tamtst
