MODULE trc_oce_tam
#if defined key_tam
   !!======================================================================
   !!                      ***  MODULE  trc_oce_tam  ***
   !! Ocean passive tracer  :  share SMS/Ocean variables
   !!                          Tangent and Adjoint Module
   !!======================================================================
   !! History of the direct module:
   !!   9.0  !  04-03  (C. Ethe)  F90: Free form and module
   !! History of the T&A module:
   !!   9.0  !  08-11  (A. Vidard)  original version
   !!----------------------------------------------------------------------
# if defined key_top && defined key_pisces
   !!----------------------------------------------------------------------
   !!  OPA 9.0 , LOCEAN-IPSL (2005) 
   !! $Id: trc_oce.F90 1152 2008-06-26 14:11:13Z rblod $ 
   !! This software is governed by the CeCILL licence see modipsl/doc/NEMO_CeCILL.txt 
   !!----------------------------------------------------------------------
   USE sms_pisces_tam , ONLY :  &
      etot3_tl    =>   etot3_tl & !!:  Biological fluxes for light
      etot3_ad    =>   etot3_ad   !!:  Biological fluxes for light
# else
   !!----------------------------------------------------------------------
   !! Default option                         No Biological fluxes for light          
   !!----------------------------------------------------------------------
   USE par_kind, ONLY : &       ! Precision variables
      & wp
   USE par_oce, ONLY: &
      & jpi,          &
      & jpj,          &
      & jpk
   !! * Module variables
   REAL(wp), DIMENSION(:,:,:), ALLOCATABLE :: &
      & etot3_tl, &
      & etot3_ad
# endif
   PUBLIC &
      & trc_oce_tam_init, & !: Initialize the trend TAM fields
      & trc_oce_tam_deallocate, & !: Deallocate the trend TAM fields
      & etot3_tl, &
      & etot3_ad

CONTAINS

   SUBROUTINE trc_oce_tam_init( kindic )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE trc_oce_tam_init  ***
      !!
      !! ** Purpose : Allocate and initialize the tangent linear and
      !!              adjoint arrays
      !!
      !! ** Method  : kindic = 0  allocate/initialize both tl and ad variables
      !!              kindic = 1  allocate/initialize only tl variables
      !!              kindic = 2  allocate/initialize only ad variables
      !!
      !! ** Action  :
      !!
      !! References :
      !!
      !! History :
      !!        ! 09-05 (F. Vigilant) Initial version
      !!-----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kindic        ! indicate which variables to allocate/initialize

      !! * Local declarations
      ! Allocate tangent linear variable arrays
      ! ---------------------------------------

      IF ( kindic == 0 .OR. kindic == 1 ) THEN

         IF ( .NOT. ALLOCATED(etot3_tl) ) THEN

            ALLOCATE( etot3_tl(jpi,jpj,jpk) )

         ENDIF

         ! Initialize tangent linear variable arrays to zero
         ! -------------------------------------------------

         etot3_tl   (:,:,:) = 0.0_wp

      END IF

      IF ( kindic == 0 .OR. kindic == 2 ) THEN
            
         ! Allocate adjoint variable arrays
         ! --------------------------------

         IF ( .NOT. ALLOCATED(etot3_ad) ) THEN

            ALLOCATE( etot3_ad(jpi,jpj,jpk) )

         ENDIF

         ! Initialize adjoint variable arrays to zero
         ! ------------------------------------------

         etot3_ad   (:,:,:) = 0.0_wp

      END IF

   END SUBROUTINE trc_oce_tam_init
   SUBROUTINE trc_oce_tam_deallocate( kindic )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE trc_oce_tam_deallocate  ***
      !!
      !! ** Purpose : Deallocate the tangent linear and
      !!              adjoint arrays
      !!
      !! ** Method  : kindic = 0  deallocate both tl and ad variables
      !!              kindic = 1  deallocate only tl variables
      !!              kindic = 2  deallocate only ad variables
      !!
      !! ** Action  :
      !!
      !! References :
      !!
      !! History :
      !!        ! 2010-06 (A. Vidard) Initial version
      !!-----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kindic        ! indicate which variables to deallocate

      !! * Local declarations
      ! Deallocate tangent linear variable arrays
      ! ---------------------------------------

      IF ( kindic == 0 .OR. kindic == 1 ) THEN

         IF ( ALLOCATED(etot3_tl) ) DEALLOCATE( etot3_tl )

      END IF

      IF ( kindic == 0 .OR. kindic == 2 ) THEN
            
         ! Deallocate adjoint variable arrays
         ! --------------------------------

         IF ( ALLOCATED(etot3_ad) ) DEALLOCATE( etot3_ad )

      END IF

   END SUBROUTINE trc_oce_tam_deallocate
#endif
END MODULE trc_oce_tam

